import abc
import json
import os
import pydoc

import six
import terminaltables
from halo import halo

from gradient import api_sdk, exceptions, JobArtifactsDownloader, cli_constants
from gradient.api_sdk import config, sdk_exceptions
from gradient.api_sdk import utils
from gradient.api_sdk.clients import http_client
from gradient.api_sdk.utils import print_dict_recursive, concatenate_urls, MultipartEncoder
from gradient.cliutils import get_terminal_lines
from gradient.commands.common import BaseCommand, StreamMetricsCommand, LogsCommandMixin


@six.add_metaclass(abc.ABCMeta)
class BaseJobCommand(BaseCommand):
    def _get_client(self, api_key, logger_):
        client = api_sdk.clients.JobsClient(
            api_key=api_key,
            logger=logger_,
            ps_client_name=cli_constants.CLI_PS_CLIENT_NAME,
        )
        return client


@six.add_metaclass(abc.ABCMeta)
class BaseCreateJobCommandMixin(BaseJobCommand):
    SPINNER_MESSAGE = "Creating new job"
    CREATE_SUCCESS_MESSAGE_TEMPLATE = "New job created with ID: {}"

    def __init__(self, workspace_handler, *args, **kwargs):
        super(BaseCreateJobCommandMixin, self).__init__(*args, **kwargs)
        self.workspace_handler = workspace_handler

    def execute(self, json_):
        json_, data = self._handle_workspace(json_)

        with halo.Halo(text=self.SPINNER_MESSAGE, spinner="dots"):
            job_id = self._create(json_, data)

        self.logger.log(self.CREATE_SUCCESS_MESSAGE_TEMPLATE.format(job_id))
        self.logger.log(self.get_instance_url(job_id))

    def get_instance_url(self, instance_id):
        url = concatenate_urls(config.config.WEB_URL, "{}/jobs/{}".format(self.get_namespace(), instance_id))
        return url

    def _handle_workspace(self, instance_dict):
        """

        :param instance_dict:
        :return:
        """
        data = None

        self._set_project_if_not_provided(instance_dict)
        workspace_url = self.workspace_handler.handle(instance_dict)
        if workspace_url:
            if utils.PathParser.is_local_zip_file(workspace_url):
                data = self._get_multipart_data(workspace_url, instance_dict)
            else:
                instance_dict["workspace_file_name"] = workspace_url

        return instance_dict, data

    def _get_multipart_data(self, workspace_url, json_):
        archive_basename = os.path.basename(workspace_url)
        json_["workspace_file_name"] = archive_basename
        job_data = self._get_files_dict(workspace_url, archive_basename)
        monitor = MultipartEncoder(job_data).get_monitor()
        return monitor

    def _get_files_dict(self, workspace_url, archive_basename):
        job_data = {'file': (archive_basename, open(workspace_url, 'rb'), 'text/plain')}
        return job_data

    @staticmethod
    def _set_project_if_not_provided(json_):
        if not json_.get("project_id"):
            json_["project"] = "gradient-project"

    @abc.abstractmethod
    def _create(self, json_, data):
        pass


class DeleteJobCommand(BaseJobCommand):

    def execute(self, job_id):
        self.client.delete(job_id)
        self.logger.log("Job {} deleted".format(job_id))


class StopJobCommand(BaseJobCommand):

    def execute(self, job_id):
        self.client.stop(job_id)
        self.logger.log("Job {} stopped".format(job_id))


class ListJobsCommand(BaseJobCommand):
    WAITING_FOR_RESPONSE_MESSAGE = "Waiting for data..."

    def execute(self, **kwargs):
        with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
            instances = self._get_instances(**kwargs)

        self._log_objects_list(instances)

    def _get_instances(self, **kwargs):

        try:
            instances = self.client.list(**kwargs)
        except sdk_exceptions.GradientSdkError as e:
            raise exceptions.ReceivingDataFailedError(e)

        return instances

    @staticmethod
    def _get_table_data(jobs):
        data = [("ID", "Name", "Project", "Cluster", "Machine Type", "Created")]
        for job in jobs:
            id_ = job.id
            name = job.name
            project = job.project
            cluster = job.cluster
            machine_type = job.machine_type
            created = job.dt_created
            data.append((id_, name, project, cluster, machine_type, created))

        return data

    def _log_objects_list(self, objects):
        if not objects:
            self.logger.warning("No data found")
            return

        table_data = self._get_table_data(objects)
        table_str = self._make_table(table_data)
        if len(table_str.splitlines()) > get_terminal_lines():
            pydoc.pager(table_str)
        else:
            self.logger.log(table_str)

    @staticmethod
    def _make_table(table_data):
        ascii_table = terminaltables.AsciiTable(table_data)
        table_string = ascii_table.table
        return table_string


class JobLogsCommand(LogsCommandMixin, BaseJobCommand):
    ENTITY = "Job"


class CreateJobCommand(BaseCreateJobCommandMixin):

    def _create(self, json_, data):
        # because ignore_files is used by workspace handlers and not needed anymore (will fail if not "popped")
        json_.pop("ignore_files", None)
        json_.pop("workspace", None)

        return self.client.create(data=data, **json_)


class ArtifactsDestroyCommand(BaseJobCommand):
    def execute(self, job_id, files=None):
        self.client.artifacts_delete(job_id, files)
        self.logger.log("Job {} artifacts deleted".format(job_id))


class ArtifactsGetCommand(BaseJobCommand):
    def execute(self, job_id):
        artifact = self.client.artifacts_get(job_id)

        self._log_artifacts(artifact, job_id)

    def _log_artifacts(self, artifact, job_id):
        if artifact:
            print_dict_recursive(artifact, self.logger)
        else:
            self.logger.log("No artifacts found for job {}".format(job_id))


class ArtifactsListCommand(BaseJobCommand):
    WAITING_FOR_RESPONSE_MESSAGE = "Waiting for data..."

    def execute(self, **kwargs):
        with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
            try:
                start_after = None
                instances = []
                while True:
                    pagination_response = self.client.artifacts_list(start_after=start_after, **kwargs)

                    if pagination_response.data:
                        instances.extend(pagination_response.data)
                    start_after = pagination_response.start_after

                    if start_after is None:
                        break
            except sdk_exceptions.GradientSdkError as e:
                raise exceptions.ReceivingDataFailedError(e)

        self._log_objects_list(instances, kwargs)

    def _get_table_data(self, artifacts, kwargs):
        columns = ['Files']

        show_size = "size" in kwargs
        show_url = "url" in kwargs

        if show_size:
            columns.append('Size (in bytes)')
        if show_url:
            columns.append('URL')

        data = [tuple(columns)]
        for artifact in artifacts:
            row = [artifact.file]
            if show_size:
                row.append(artifact.size)
            if show_url:
                row.append(artifact.url)
            data.append(tuple(row))
        return data

    def _log_objects_list(self, objects, kwargs):
        if not objects:
            self.logger.warning("No data found")
            return

        table_data = self._get_table_data(objects, kwargs)
        table_str = self._make_table(table_data)
        if len(table_str.splitlines()) > get_terminal_lines():
            pydoc.pager(table_str)
        else:
            self.logger.log(table_str)

    @staticmethod
    def _make_table(table_data):
        ascii_table = terminaltables.AsciiTable(table_data)
        table_string = ascii_table.table
        return table_string


class DownloadArtifactsCommand(BaseJobCommand):
    WAITING_FOR_RESPONSE_MESSAGE = "Waiting for data..."

    def execute(self, job_id, destination_directory):
        artifact_downloader = JobArtifactsDownloader(
            self.api_key,
            logger=self.logger,
            ps_client_name=cli_constants.CLI_PS_CLIENT_NAME,
        )
        with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
            try:
                artifact_downloader.download(job_id, destination_directory)
            except OSError as e:
                raise sdk_exceptions.GradientSdkError(e)


class JobAddTagsCommand(BaseJobCommand):
    def execute(self, job_id, *args, **kwargs):
        self.client.add_tags(job_id, **kwargs)
        self.logger.log("Tags added to job")


class JobRemoveTagsCommand(BaseJobCommand):
    def execute(self, job_id, *args, **kwargs):
        self.client.remove_tags(job_id, **kwargs)
        self.logger.log("Tags removed from job")


class GetJobMetricsCommand(BaseJobCommand):
    def execute(self, deployment_id, start, end, interval, built_in_metrics, *args, **kwargs):
        metrics = self.client.get_metrics(
            deployment_id,
            start=start,
            end=end,
            built_in_metrics=built_in_metrics,
            interval=interval,
        )
        formatted_metrics = json.dumps(metrics, indent=2, sort_keys=True)
        self.logger.log(formatted_metrics)


class ListJobMetricsCommand(BaseJobCommand):
    def execute(self, deployment_id, start, end, interval, *args, **kwargs):
        metrics = self.client.list_metrics(
            deployment_id,
            start=start,
            end=end,
            interval=interval,
        )
        formatted_metrics = json.dumps(metrics, indent=2, sort_keys=True)
        self.logger.log(formatted_metrics)


class StreamJobMetricsCommand(StreamMetricsCommand, BaseJobCommand):
    pass



from operator import itemgetter

from .. import config, serializers
from ..repositories.common import ListResources


class ListMachineTypes(ListResources):
    SERIALIZER_CLS = serializers.VmTypeSchema

    def get_request_url(self, **kwargs):
        return "vmTypes/getVmTypesByClusters"

    def _get_api_url(self, **kwargs):
        return config.config.CONFIG_HOST

    def _get_instance_dicts(self, data, cluster_id=None, **kwargs):
        vm_types_dicts = {}  # vmType["label"]: vmType dict
        for cluster_list_of_vms in data.values():
            current_cluster_id = cluster_list_of_vms[0]["clusters"][0]["id"]

            for vm in cluster_list_of_vms:
                vm_type = vm["vmType"]
                vm_type_label = vm_type["label"]
                vm_types_dicts.setdefault(vm_type_label, vm_type)

                clusters = vm_types_dicts[vm_type_label].setdefault("clusters", [])
                if current_cluster_id not in clusters:
                    clusters.append(current_cluster_id)

        vm_types = list(vm_types_dicts.values())
        if cluster_id is not None:
            vm_types = [vm_type for vm_type in vm_types
                        if cluster_id in vm_type["clusters"]]

        vm_types.sort(key=itemgetter("label"))
        for vm_type in vm_types:
            vm_type["clusters"].sort()

        return vm_types



import marshmallow as ma

from .base import BaseSchema
from .tag import TagSchema
from .. import models


class VolumeOptionsSchema(BaseSchema):
    MODEL = models.VolumeOptions

    kind = ma.fields.Str(required=True)
    size = ma.fields.Str(required=True)


class ExperimentDatasetSchema(BaseSchema):
    MODEL = models.ExperimentDataset

    id = ma.fields.String()
    uri = ma.fields.String()
    aws_access_key_id = ma.fields.String(dump_to="awsAccessKeyId", load_from="awsAccessKeyId")
    aws_secret_access_key = ma.fields.String(dump_to="awsSecretAccessKey", load_from="awsSecretAccessKey")
    aws_endpoint = ma.fields.String(dump_to="awsEndpoint", load_from="awsEndpoint")
    etag = ma.fields.String()
    version_id = ma.fields.String(dump_to="versionId", load_from="versionId")
    name = ma.fields.String()
    tags = ma.fields.Nested(TagSchema, only="name", many=True, load_only=True)
    volume_options = ma.fields.Nested(VolumeOptionsSchema, dump_to="volumeOptions", load_from="volumeOptions")



import os
import shutil
import tempfile
import zipfile

import mock

import gradient.api_sdk.archivers
import gradient.api_sdk.s3_uploader


def create_file(dir_path, filename):
    file_path = os.path.join(dir_path, filename)
    with open(file_path, "w") as h:
        h.write("keton")


def create_test_dir_tree(dir_path=None):
    if dir_path is None:
        dir_path = tempfile.mkdtemp()

    create_file(dir_path, "file1.txt")
    create_file(dir_path, "file2.jpg")

    dot_git_path = os.path.join(dir_path, ".git")
    os.mkdir(dot_git_path)
    create_file(dot_git_path, "some_file")

    subdir_path = os.path.join(dir_path, "subdir1")
    os.mkdir(subdir_path)
    create_file(subdir_path, "file2.jpg")
    create_file(subdir_path, "file3.txt")

    subdir2_path = os.path.join(dir_path, "subdir2")
    os.mkdir(subdir2_path)
    create_file(subdir2_path, "file4")
    subdir21_path = os.path.join(subdir2_path, "subdir21")
    os.mkdir(subdir21_path)
    create_file(subdir21_path, "file5")

    subdir3_path = os.path.join(dir_path, "subdir3")
    os.mkdir(subdir3_path)
    create_file(subdir3_path, "file4")
    subdir31_path = os.path.join(subdir3_path, "subdir31")
    os.mkdir(subdir31_path)
    create_file(subdir31_path, "file5")

    return dir_path


class TestZipArchiver(object):
    def test_should_get_valid_excluded_paths_when_empty_list_was_passed_to_get_excluded_paths(self):
        archiver = gradient.api_sdk.archivers.ZipArchiver()

        excluded = archiver.get_excluded_paths()

        assert excluded == {
            os.path.join(".git", "*"),
            os.path.join(".idea", "*"),
            os.path.join(".pytest_cache", "*"),
        }

    def test_should_get_valid_excluded_paths_when_list_of_files_was_passed_to_get_excluded_paths(self):
        archiver = gradient.api_sdk.archivers.ZipArchiver()

        excluded = archiver.get_excluded_paths(["some_file", "some_dir/some_other_file"])

        assert excluded == {
            os.path.join(".git", "*"),
            os.path.join(".idea", "*"),
            os.path.join(".pytest_cache", "*"),
            "some_file",
            os.path.join("some_dir", "some_other_file"),
        }

    def test_should_get_a_dictionary_of_file_paths_in_a_dir(self):
        test_dir = create_test_dir_tree()
        expected_paths = {
            "file1.txt": os.path.join(test_dir, "file1.txt"),
            "file2.jpg": os.path.join(test_dir, "file2.jpg"),
            os.path.join(".git", "some_file"): os.path.join(test_dir, ".git", "some_file"),
            os.path.join("subdir1", "file2.jpg"): os.path.join(test_dir, "subdir1", "file2.jpg"),
            os.path.join("subdir1", "file3.txt"): os.path.join(test_dir, "subdir1", "file3.txt"),
            os.path.join("subdir2", "subdir21", "file5"): os.path.join(test_dir, "subdir2", "subdir21", "file5"),
        }

        try:
            paths = gradient.api_sdk.archivers.ZipArchiver.get_file_paths(
                test_dir,
                excluded_paths=["file2",
                                os.path.join("subdir2", "file4"),
                                os.path.join("subdir3", "*"),
                                ],
            )

            assert paths == expected_paths

        finally:
            shutil.rmtree(test_dir)

    def test_should_add_files_to_archive_when_zip_archiver_was_used(self):
        temp_dir = tempfile.mkdtemp()
        temp_input_dir = os.path.join(temp_dir, "temp_dir")
        os.mkdir(temp_input_dir)
        create_test_dir_tree(temp_input_dir)

        archive_file_path = os.path.join(temp_dir, "archive.zip")
        create_file(temp_dir, "archive.zip")

        temp_dir_for_extracted_files = os.path.join(temp_dir, "output_dir")
        os.mkdir(temp_dir_for_extracted_files)

        expected_paths = {
            "file1.txt",
            "file2.jpg",
            os.path.join("subdir1", "file2.jpg"),
            os.path.join("subdir1", "file3.txt"),
            os.path.join("subdir2", "subdir21", "file5"),
        }

        try:
            archiver = gradient.api_sdk.archivers.ZipArchiver()
            archiver.archive(temp_input_dir, archive_file_path,
                             exclude=["file2", os.path.join("subdir2", "file4"), os.path.join("subdir3", "*")])

            zip_file = zipfile.ZipFile(archive_file_path)
            zip_file.extractall(temp_dir_for_extracted_files)
            archiver2 = gradient.api_sdk.archivers.ZipArchiver()
            archiver2.default_excluded_paths = []
            paths_in_extracted_dir = archiver2.get_file_paths(temp_dir_for_extracted_files)

        finally:
            shutil.rmtree(temp_dir)

        assert set(paths_in_extracted_dir.keys()) == expected_paths


class TestS3FileUploader(object):
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_upload_file_to_s3_and_get_bucket_url_when_upload_was_executed(self, post_patched):
        _, file_path = tempfile.mkstemp()
        uploader = gradient.api_sdk.s3_uploader.S3FileUploader()

        uploader.upload(file_path, "s3://some.url", {"key": "some_key"})

        post_patched.assert_called_once()


class TestExperimentWorkspaceDirectoryUploader(object):
    WORKSPACE_DIR_PATH = "/some/workspace/dir/path/"
    TEMP_DIR_PATH = "/some/temp/dir/path/"

    @mock.patch("gradient.api_sdk.s3_uploader.ExperimentFileUploader")
    @mock.patch("gradient.api_sdk.s3_uploader.ZipArchiver")
    def test_class_with_default_params(self, zip_archiver_cls, s3_project_file_uploader_cls):
        zip_archiver = mock.MagicMock()
        zip_archiver_cls.return_value = zip_archiver
        s3_project_file_uploader = mock.MagicMock()
        s3_project_file_uploader.upload.return_value = "s3://url/to/bucket"
        s3_project_file_uploader_cls.return_value = s3_project_file_uploader
        archive_path = os.path.join(tempfile.gettempdir(), "temp.zip")

        uploader = gradient.ExperimentWorkspaceDirectoryUploader("some_api_key", ps_client_name="some_client_name")
        bucket_url = uploader.upload(self.WORKSPACE_DIR_PATH, "some_project_id")

        s3_project_file_uploader_cls.assert_called_once_with("some_api_key", ps_client_name="some_client_name")
        zip_archiver.archive.assert_called_once_with(self.WORKSPACE_DIR_PATH, archive_path, exclude=None)
        s3_project_file_uploader.upload.assert_called_once_with(archive_path, "some_project_id")
        assert bucket_url == "s3://url/to/bucket"

    def test_should_run_upload_(self):
        zip_archiver = mock.MagicMock()
        s3_project_file_uploader = mock.MagicMock()
        s3_project_file_uploader.upload.return_value = "s3://url/to/bucket"
        temp_file_name = "some_temp_file_name.zip"
        archive_path = os.path.join(self.TEMP_DIR_PATH, temp_file_name)

        uploader = gradient.ExperimentWorkspaceDirectoryUploader(
            "some_api_key",
            temp_dir=self.TEMP_DIR_PATH,
            archiver=zip_archiver,
            project_uploader=s3_project_file_uploader,
        )
        bucket_url = uploader.upload(
            self.WORKSPACE_DIR_PATH,
            "some_project_id",
            exclude=["file1", "dir/file2"],
            temp_file_name=temp_file_name,
        )

        zip_archiver.archive.assert_called_once_with(
            self.WORKSPACE_DIR_PATH,
            os.path.join(self.TEMP_DIR_PATH, temp_file_name),
            exclude=["file1", "dir/file2"],
        )
        s3_project_file_uploader.upload.assert_called_once_with(archive_path, "some_project_id")
        assert bucket_url == "s3://url/to/bucket"



import abc

import six


@six.add_metaclass(abc.ABCMeta)
class Logger(object):
    @abc.abstractmethod
    def log(self, msg, *args, **kwargs):
        pass

    @abc.abstractmethod
    def warning(self, msg, *args, **kwargs):
        pass

    @abc.abstractmethod
    def error(self, msg, *args, **kwargs):
        pass

    def debug(self, msg, *args, **kwargs):
        pass


class MuteLogger(Logger):
    def log(self, msg, *args, **kwargs):
        pass

    def warning(self, msg, *args, **kwargs):
        pass

    def error(self, msg, *args, **kwargs):
        pass



"""
Tensorboard logic related client handler.

Remember that in code snippets all highlighted lines are required other lines are optional.
"""
from .base_client import BaseClient
from .. import repositories, models


class TensorboardClient(BaseClient):
    """
    Client to handle tensorboard related actions.

    How to create instance of tensorboard client:

    .. code-block:: python
        :linenos:
        :emphasize-lines: 4

        from gradient import TensorboardClient

        tb_client = TensorboardClient(
            api_key='your_api_key_here'
        )

    """

    def create(
            self,
            image=None,
            username=None,
            password=None,
            instance_type=None,
            instance_size=None,
            instances_count=None,
            experiments=None,
    ):
        """
        Method to create tensorboard in paperspace gradient.

        Example create tensorboard:

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2

            tb_id = tb_client.create(
                experiments=['some_experiment_id'],
                image='tensorflow/tensorflow:latest-py3',
                username='your_username',
                password='your_password',
                instance_type='cpu',
                instance_size='small',
                instance_count=1
            )

        :param str image: your tensorboard will run with this image.
            By default it will be run with ``tensorflow/tensorflow:latest-py3``
        :param str username: if you wish to limit access to your tensorboard with base auth then provide username
        :param str password: if you wish to limit access to your tensorboard with base auth then provide password
        :param str instance_type: type of instance on which you want to run tensorboard.
            Available choices:

            .. code-block::

                cpu
                gpu

            By default we use ``cpu`` instance type.
        :param str instance_size: size of instance on which you want to run tensorboard.
            Available choices:

            .. code-block::

                small
                medium
                large

            By default we use ``small`` instance size.
        :param int instances_count: on how many machines you want to run tensorboard. By default ``1`` is used.
        :param list experiments: list of experiments that you wish to add to tensorboard.
            To create tensorboard you need to provide at least one experiment id. This field is **required**.

        :return: Return tensorboard id
        :rtype: str

        :raises: ResourceFetchingError: When there is problem with response from API
        """
        tensorboard = models.Tensorboard(
            image=image,
            username=username,
            password=password,
            instance=models.Instance(
                type=instance_type,
                size=instance_size,
                count=instances_count,
            ),
            experiments=experiments,
        )

        repository = self.build_repository(repositories.CreateTensorboard)
        tensorboard_id = repository.create(tensorboard)
        return tensorboard_id

    def get(self, id):
        """
        Method to get tensorboard details.

        Example get tensorboard details:

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2

            tb = tb_client.get(
                id='your_tb_id'
            )

        :param str id: Tensorboard id of which you want to get details
        :return: Tensorbord object if found
        :rtype: None|Tensorboard

        :raises: ResourceFetchingError: When there is problem with response from API
        """
        repository = self.build_repository(repositories.GetTensorboard)
        tensorboard = repository.get(id=id)
        return tensorboard

    def list(self):
        """
        Method to list your active tensorboards.

        Example usage:

        .. code-block:: python
            :linenos:

            tb_list = tb_client.list()

        :return: list of active tensorboards
        :rtype: list[models.Tensorboard]

        :raises: ResourceFetchingError: When there is problem with response from API
        """
        repository = self.build_repository(repositories.ListTensorboards)
        tensorboards = repository.list()
        return tensorboards

    def add_experiments(self, id, added_experiments):
        """
        Method to add experiments to existing tensorboard.

        Example usage:

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2, 3

            tb = tb_client.add_experiments(
                id='your_tb_id',
                added_experiments=['new_experiment_id', 'next_new_experiment_id']
            )

        :param str id: tensorboard id to which you want to add experiments
        :param list added_experiments: list of experiment ids which you want to add to tensroboard

        :return: updated tensorboard
        :rtype: Tensorboard

        :raises: ResourceFetchingError: When there is problem with response from API
        """
        repository = self.build_repository(repositories.UpdateTensorboard)
        tensorboard = repository.update(id=id, added_experiments=added_experiments)
        return tensorboard

    def remove_experiments(self, id, removed_experiments):
        """
        Method to remove experiments from existing tensorboard.

        Example usage:

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2, 3

            tb = tb_client.remove_experiments(
                id='your_tb_id',
                removed_experiments=['experiment_id', 'next_experiment_id']
            )

        :param str id: tensorboard id from which you want to remove experiments
        :param list removed_experiments: list of experiment ids which you want to remove from tensroboard

        :return: updated tensorboard
        :rtype: Tensorboard

        :raises: ResourceFetchingError: When there is problem with response from API
        """
        repository = self.build_repository(repositories.UpdateTensorboard)
        tensorboard = repository.update(id=id, removed_experiments=removed_experiments)
        return tensorboard

    def delete(self, id):
        """
        Method to delete tensorboard.

        Example usage:

        .. code-block:: python
            :linenos:
            :emphasize-lines: 2

            tb_client.delete(
                id='your_tb_id'
            )

        :param str id: Tensoboard id which you want to delete
        """
        repository = self.build_repository(repositories.DeleteTensorboard)
        repository.delete(id_=id)



import collections

import click

from gradient import clilogger
from gradient.api_sdk import constants
from gradient.cli import common
from gradient.cli.cli import cli
from gradient.cli.cli_types import ChoiceType, json_string
from gradient.cli.common import api_key_option, del_if_value_is_none, ClickGroup, validate_comma_split_option
from gradient.cli.validators import validate_email, validate_mutually_exclusive
from gradient.commands import machines as machines_commands

REGIONS_MAP = collections.OrderedDict(
    (
        ("CA1", constants.Region.CA1),
        ("NY2", constants.Region.NY2),
        ("AMS1", constants.Region.AMS1),
    )
)


@cli.group("machines", help="Manage machines", cls=ClickGroup)
def machines_group():
    pass


@machines_group.group("tags", help="Manage machine tags", cls=ClickGroup)
def machines_tags():
    pass


check_machine_availability_help = "Get machine availability for the given region and machine type. " \
                                  "Note: availability is only provided for the dedicated GPU machine types. " \
                                  "Also, not all machine types are available in all regions"


@machines_group.command("availability", help=check_machine_availability_help)
@click.option(
    "--region",
    "region",
    type=ChoiceType(REGIONS_MAP, case_sensitive=False),
    required=True,
    help="Name of the region",
    cls=common.GradientOption,
)
@click.option(
    "--machineType",
    "machine_type",
    required=True,
    help="Machine type",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def check_machine_availability(region, machine_type, api_key, options_file):
    command = machines_commands.CheckAvailabilityCommand(api_key=api_key)
    command.execute(machine_type, region)


create_machine_help = "Create a new Paperspace virtual machine. If you are using an individual account, you will " \
                      "be assigned as the owner of the machine. If you are a team administrator, you must specify " \
                      "the user that should be assigned to the machine, either by specifing a user id, or by " \
                      "providing an email address, password, first name and last name for the creation of a new " \
                      "user on the team."


@machines_group.command("create", help=create_machine_help)
@click.option(
    "--region",
    "region",
    type=ChoiceType(REGIONS_MAP, case_sensitive=False),
    required=True,
    help="Name of the region",
    cls=common.GradientOption,
)
@click.option(
    "--machineType",
    "machine_type",
    required=True,
    help="Machine type",
    cls=common.GradientOption,
)
@click.option(
    "--size",
    "size",
    type=int,
    required=True,
    help="Storage size for the machine in GB",
    cls=common.GradientOption,
)
@click.option(
    "--billingType",
    "billing_type",
    type=click.Choice(constants.BILLING_TYPES),
    required=True,
    help="Either 'monthly' or 'hourly' billing",
    cls=common.GradientOption,
)
@click.option(
    "--machineName",
    "name",
    required=True,
    help="A memorable name for this machine",
    cls=common.GradientOption,
)
@click.option(
    "--templateId",
    "template_id",
    required=True,
    help="Template id of the template to use for creating this machine",
    cls=common.GradientOption,
)
@click.option(
    "--assignPublicIp",
    "assign_public_ip",
    is_flag=True,
    default=None,  # None is used so it can be filtered with `del_if_value_is_none` when flag was not set
    help="Assign a new public ip address on machine creation. Cannot be used with dynamicPublicIp",
    cls=common.GradientOption,
)
@click.option(
    "--dynamicPublicIp",
    "dynamic_public_ip",
    is_flag=True,
    default=None,  # None is used so it can be filtered with `del_if_value_is_none` when flag was not set
    help="Assigns a new public ip address on machine start and releases it from the account on machine stop. "
         "Cannot be used with assignPublicIp",
    cls=common.GradientOption,
)
@click.option(
    "--networkId",
    "network_id",
    help="If creating on a specific network, specify its id",
    cls=common.GradientOption,
)
@click.option(
    "--teamId",
    "team_id",
    help="If creating the machine for a team, specify the team id",
    cls=common.GradientOption,
)
@click.option(
    "--userId",
    "user_id",
    help="If assigning to an existing user other than yourself, specify the user id (mutually exclusive with email, "
         "password, firstName, lastName)",
    cls=common.GradientOption,
)
@click.option(
    "--email",
    "email",
    help="If creating a new user for this machine, specify their email address (mutually exclusive with userId)",
    callback=validate_email,
    cls=common.GradientOption,
)
@click.option(
    "--password",
    "password",
    help="If creating a new user, specify their password (mutually exclusive with userId)",
    cls=common.GradientOption,
)
@click.option(
    "--firstName",
    "first_name",
    help="If creating a new user, specify their first name (mutually exclusive with userId)",
    cls=common.GradientOption,
)
@click.option(
    "--lastName",
    "last_name",
    help="If creating a new user, specify their last name (mutually exclusive with userId)",
    cls=common.GradientOption,
)
@click.option(
    "--notificationEmail",
    "notification_email",
    help="Send a notification to this email address when complete",
    callback=validate_email,
    cls=common.GradientOption,
)
@click.option(
    "--scriptId",
    "script_id",
    help="The script id of a script to be run on startup. See the Script Guide for more info on using scripts",
    cls=common.GradientOption,
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to add to experiment",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want add to experiment",
    cls=common.GradientOption
)
@api_key_option
@common.options_file
def create_machine(api_key, options_file, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"))
    del_if_value_is_none(kwargs)

    assign_public_ip = kwargs.get("assign_public_ip")
    dynamic_public_ip = kwargs.get("dynamic_public_ip")
    validate_mutually_exclusive([assign_public_ip], [dynamic_public_ip],
                                "--assignPublicIp cannot be used with --dynamicPublicIp")

    user_id = kwargs.get("user_id")
    email = kwargs.get("email")
    password = kwargs.get("password")
    first_name = kwargs.get("first_name")
    last_name = kwargs.get("last_name")
    validate_mutually_exclusive([user_id], [email, password, first_name, last_name],
                                "--userId is mutually exclusive with --email, --password, --firstName and --lastName")

    command = machines_commands.CreateMachineCommand(api_key=api_key, logger=clilogger.CliLogger())
    command.execute(kwargs)


destroy_machine_help = "Destroy the machine with the given id. When this action is performed, the machine is " \
                       "immediately shut down and marked for deletion from the datacenter. Any snapshots that " \
                       "were derived from the machine are also deleted. Access to the machine is terminated " \
                       "immediately and billing for the machine is prorated to the hour. This action can only " \
                       "be performed by the user who owns the machine, or in the case of a team, the team " \
                       "administrator."


@machines_group.command("destroy", help=destroy_machine_help)
@click.option(
    "--id",
    "machine_id",
    required=True,
    help="The id of the machine to destroy",
    cls=common.GradientOption,
)
@click.option(
    "--releasePublicIp",
    "release_public_ip",
    is_flag=True,
    help="releases any assigned public ip address for the machine; defaults to false",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def destroy_machine(machine_id, release_public_ip, api_key, options_file):
    command = machines_commands.DestroyMachineCommand(api_key=api_key, logger=clilogger.CliLogger())
    command.execute(machine_id, release_public_ip)


list_machines_help = "List information about all machines available to either the current authenticated user or " \
                     "the team, if the user belongs to a team. The list method takes an optional first argument " \
                     "to limit the returned machine objects."


@machines_group.command("list", help=list_machines_help)
@click.option(
    "--params",
    "params",
    type=json_string,
    help="JSON used to filter machines. Use either this or a combination of following options",
    cls=common.GradientOption,
)
@click.option(
    "--id",
    "id",
    help="Optional machine id to match on",
    cls=common.GradientOption,
)
@click.option(
    "--name",
    "name",
    help="Filter by machine name",
    cls=common.GradientOption,
)
@click.option(
    "--os",
    "os",
    help="Filter by os used",
    cls=common.GradientOption,
)
@click.option(
    "--ram",
    "ram",
    type=int,
    help="Filter by machine RAM (in bytes)",
    cls=common.GradientOption,
)
@click.option(
    "--cpus",
    "cpus",
    type=int,
    help="Filter by CPU count",
    cls=common.GradientOption,
)
@click.option(
    "--gpu",
    "gpu",
    help="Filter by GPU type",
    cls=common.GradientOption,
)
@click.option(
    "--storageTotal",
    "storage_total",
    help="Filter by total storage",
    cls=common.GradientOption,
)
@click.option(
    "--storageUsed",
    "storage_used",
    help="Filter by storage used",
    cls=common.GradientOption,
)
@click.option(
    "--usageRate",
    "usage_rate",
    help="Filter by usage rate",
    cls=common.GradientOption,
)
@click.option(
    "--shutdownTimeoutInHours",
    "shutdown_timeout_in_hours",
    type=int,
    help="Filter by shutdown timeout",
    cls=common.GradientOption,
)
@click.option(
    "--performAutoSnapshot",
    "perform_auto_snapshot",
    type=bool,
    help="Filter by performAutoSnapshot flag",
    cls=common.GradientOption,
)
@click.option(
    "--autoSnapshotFrequency",
    "auto_snapshot_frequency",
    type=click.Choice(["hour", "day", "week"], case_sensitive=False),
    help="Filter by autoSnapshotFrequency flag",
    cls=common.GradientOption,
)
@click.option(
    "--autoSnapshotSaveCount",
    "auto_snapshot_save_count",
    type=int,
    help="Filter by auto shapshots count",
    cls=common.GradientOption,
)
@click.option(
    "--agentType",
    "agent_type",
    help="Filter by agent type",
    cls=common.GradientOption,
)
@click.option(
    "--dtCreated",
    "created_timestamp",
    help="Filter by date created",
    cls=common.GradientOption,
)
@click.option(
    "--state",
    "state",
    help="Filter by state",
    cls=common.GradientOption,
)
@click.option(
    "--updatesPending",
    "updates_pending",
    help="Filter by updatesPending",
    cls=common.GradientOption,
)
@click.option(
    "--networkId",
    "network_id",
    help="Filter by network ID",
    cls=common.GradientOption,
)
@click.option(
    "--privateIpAddress",
    "private_ip_address",
    help="Filter by private IP address",
    cls=common.GradientOption,
)
@click.option(
    "--publicIpAddress",
    "public_ip_address",
    help="Filter by public IP address",
    cls=common.GradientOption,
)
@click.option(
    "--region",
    "region",
    type=ChoiceType(REGIONS_MAP, case_sensitive=False),
    help="Filter by region",
    cls=common.GradientOption,
)
@click.option(
    "--userId",
    "user_id",
    help="Filter by user ID",
    cls=common.GradientOption,
)
@click.option(
    "--teamId",
    "team_id",
    help="Filter by team ID",
    cls=common.GradientOption,
)
@click.option(
    "--dtLastRun",
    "last_run_timestamp",
    help="Filter by last run date",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def list_machines(api_key, params, options_file, **kwargs):
    del_if_value_is_none(kwargs)
    params = params or {}
    kwargs = kwargs or {}
    validate_mutually_exclusive(params.values(), kwargs.values(),
                                "You can use either --params dictionary or single filter arguments")

    command = machines_commands.ListMachinesCommand(api_key=api_key, logger=clilogger.CliLogger())
    filters = params or kwargs
    command.execute(**filters)


restart_machine_help = "Restart an individual machine. If the machine is already restarting, this action will " \
                       "request the machine be restarted again. This action can only be performed by the user " \
                       "who owns the machine"


@machines_group.command("restart", help=restart_machine_help)
@click.option(
    "--id",
    "machine_id",
    help="Id of the machine to restart",
    required=True,
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def restart_machine(machine_id, api_key, options_file):
    command = machines_commands.RestartMachineCommand(api_key=api_key, logger=clilogger.CliLogger())
    command.execute(machine_id)


show_machine_details_help = "Show machine information for the machine with the given id."


@machines_group.command("details", help=show_machine_details_help)
@click.option(
    "--id",
    "machine_id",
    help="Id of the machine to show",
    required=True,
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def show_machine_details(machine_id, api_key, options_file):
    command = machines_commands.ShowMachineCommand(api_key=api_key)
    command.execute(machine_id)


update_machine_help = "Update attributes of a machine"


@machines_group.command("update", help=update_machine_help)
@click.option(
    "--id",
    "machine_id",
    help="Id of the machine to update",
    required=True,
    cls=common.GradientOption,
)
@click.option(
    "--machineName",
    "name",
    help="New name for the machine",
    cls=common.GradientOption,
)
@click.option(
    "--shutdownTimeoutInHours",
    "shutdown_timeout_in_hours",
    help="Number of hours before machine is shutdown if no one is logged in via the Paperspace client",
    type=int,
    cls=common.GradientOption,
)
@click.option(
    "--shutdownTimeoutForces",
    "shutdown_timeout_forces",
    help="Force shutdown at shutdown timeout, even if there is a Paperspace client connection",
    type=bool,
    cls=common.GradientOption,
)
@click.option(
    "--performAutoSnapshot",
    "perform_auto_snapshot",
    help="Perform auto snapshots",
    type=bool,
    cls=common.GradientOption,
)
@click.option(
    "--autoSnapshotFrequency",
    "auto_snapshot_frequency",
    help="One of 'hour', 'day', 'week', or null",
    type=click.Choice(["hour", "day", "week"], case_sensitive=False),
    cls=common.GradientOption,
)
@click.option(
    "--autoSnapshotSaveCount",
    "auto_snapshot_save_count",
    help="Number of snapshots to save",
    type=int,
    cls=common.GradientOption,
)
@click.option(
    "--dynamicPublicIp",
    "dynamic_public_ip",
    help="If true, assigns a new public ip address on machine start and releases it from the account on machine stop",
    type=bool,
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def update_machine(machine_id, api_key, options_file, **kwargs):
    del_if_value_is_none(kwargs)
    command = machines_commands.UpdateMachineCommand(api_key=api_key)
    command.execute(machine_id, kwargs)


start_machine_help = "Start up an individual machine. If the machine is already started, this action is a no-op. " \
                     "If the machine is off, it will be booted up. This action can only be performed by the user " \
                     "who owns the machine"


@machines_group.command("start", help=start_machine_help)
@click.option(
    "--id",
    "machine_id",
    help="Id of the machine to start",
    required=True,
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def start_machine(machine_id, api_key, options_file):
    command = machines_commands.StartMachineCommand(api_key=api_key)
    command.execute(machine_id)


stop_machine_help = "Stop an individual machine. If the machine is already stopped or has been shut down, this " \
                    "action is a no-op. If the machine is running, it will be stopped and any users logged in " \
                    "will be immediately kicked out. This action can only be performed by the user who owns the machine"


@machines_group.command("stop", help=stop_machine_help)
@click.option(
    "--id",
    "machine_id",
    help="Id of the machine to start",
    required=True,
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def stop_machine(machine_id, api_key, options_file):
    command = machines_commands.StopMachineCommand(api_key=api_key)
    command.execute(machine_id)


show_machine_utilization_help = "Get machine utilization data for the machine with the given id. Machine upgrades " \
                                "are not represented in utilization data"


@machines_group.command("utilization", help=show_machine_utilization_help)
@click.option(
    "--id",
    "machine_id",
    help="Id of the machine to start",
    required=True,
    cls=common.GradientOption,
)
@click.option(
    "--billingMonth",
    "billing_month",
    required=True,
    help="Month in YYYY-MM format",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def show_machine_utilization(machine_id, billing_month, api_key, options_file):
    command = machines_commands.ShowMachineUtilizationCommand(api_key=api_key)
    command.execute(machine_id, billing_month)


wait_for_machine_state_help = "Wait for the machine with the given id to enter a certain machine state. " \
                              "This action polls the server and returns only when we detect that the machine " \
                              "has transitioned into the given state."


@machines_group.command("waitfor", help=wait_for_machine_state_help)
@click.option(
    "--id",
    "machine_id",
    help="Id of the machine to start",
    required=True,
    cls=common.GradientOption,
)
@click.option(
    "--state",
    "state",
    help="Name of the state to wait for",
    type=click.Choice(["off", "serviceready", "ready"], case_sensitive=False),
    required=True,
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def wait_for_machine_state(machine_id, state, api_key, options_file):
    command = machines_commands.WaitForMachineStateCommand(api_key=api_key)
    command.execute(machine_id, state)


@machines_tags.command("add", help="Add tags to machine")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the machine",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to add to machine",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want add to machine",
    cls=common.GradientOption
)
@api_key_option
@common.options_file
def machine_add_tag(id, options_file, api_key, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"), raise_if_no_values=True)

    command = machines_commands.MachineAddTagsCommand(api_key=api_key)
    command.execute(id, **kwargs)


@machines_tags.command("remove", help="Remove tags from machine")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the machine",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to remove from machine",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want to remove from machine",
    cls=common.GradientOption
)
@api_key_option
@common.options_file
def machine_remove_tags(id, options_file, api_key, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"), raise_if_no_values=True)

    command = machines_commands.MachineRemoveTagsCommand(api_key=api_key)
    command.execute(id, **kwargs)



import click

from gradient.cli import common
from gradient.cli.clusters import clusters
from gradient.cli.common import api_key_option
from gradient.commands.machine_types import ListMachineTypesCommand


@clusters.group("machineTypes", help="Manage machine types")
def machine_types_group():
    pass


@machine_types_group.command("list", help="List available machine types")
@click.option(
    "--clusterId",
    "cluster_id",
    help="Filter machine types by cluster ID",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def list_machine_types(cluster_id=None, options_file=None, api_key=None):
    command = ListMachineTypesCommand(api_key=api_key)
    command.execute(cluster_id=cluster_id)



import click

from gradient.cli import common
from gradient.cli.cli import cli
from gradient.cli.common import ClickGroup, api_key_option
from gradient.commands import storage_providers as commands


@cli.group("storageProviders", help="Manage storage providers", cls=ClickGroup)
def storage_providers():
    pass


@storage_providers.group("create", help="Create storage providers", cls=ClickGroup)
def create():
    pass


@storage_providers.group("update", help="Update storage providers", cls=ClickGroup)
def update():
    pass


@storage_providers.command("list", help="List storage providers")
@api_key_option
@common.options_file
def get_storage_providers_list(api_key, options_file):
    command = commands.ListStorageProvidersCommand(api_key=api_key)
    for has_more in command.execute():
        if has_more:
            click.echo("")
            click.confirm("Do you want to continue?", abort=True)
            click.echo("")


@storage_providers.command("details", help="Show storage provider details")
@click.option(
    "--id",
    "id",
    help="Storage provider ID",
    cls=common.GradientOption,
    required=True,
)
@common.api_key_option
@common.options_file
def show_storage_provider(id, api_key, options_file):
    command = commands.ShowStorageProviderDetailsCommand(api_key=api_key)
    command.execute(id)


@create.command("s3", help="Create S3 storage provider")
@click.option(
    "--name",
    "name",
    help="Storage provider name",
    cls=common.GradientOption,
    required=True,
)
@click.option(
    "--bucket",
    "bucket",
    help="S3 bucket",
    cls=common.GradientOption,
    required=True,
)
@click.option(
    "--accessKey",
    "access_key",
    help="S3 access key ID",
    cls=common.GradientOption,
    required=True,
)
@click.option(
    "--secretAccessKey",
    "secret_access_key",
    help="S3 secret access key",
    cls=common.GradientOption,
    required=True,
)
@click.option(
    "--endpoint",
    "endpoint",
    help="S3 endpoint URL",
    cls=common.GradientOption,
)
@click.option(
    "--region",
    "region",
    help="S3 region",
    cls=common.GradientOption,
)
@click.option(
    "--signatureVersion",
    "signature_version",
    help="S3 signature version (ex: v4)",
    cls=common.GradientOption,
)
@common.api_key_option
@common.options_file
def create_s3_storage_provider(
        name,
        bucket,
        access_key,
        secret_access_key,
        endpoint,
        region,
        signature_version,
        api_key,
        options_file,
):
    command = commands.CreateS3StorageProviderCommand(api_key=api_key)
    command.execute(
        name=name,
        bucket=bucket,
        access_key=access_key,
        secret_access_key=secret_access_key,
        endpoint=endpoint,
        region=region,
        signature_version=signature_version,
    )


@update.command("s3", help="Update S3 storage provider")
@click.option(
    "--id",
    "id",
    help="Storage provider ID",
    cls=common.GradientOption,
    required=True,
)
@click.option(
    "--name",
    "name",
    help="Storage provider name",
    cls=common.GradientOption,
)
@click.option(
    "--bucket",
    "bucket",
    help="S3 bucket",
    cls=common.GradientOption,
)
@click.option(
    "--accessKey",
    "access_key",
    help="S3 access key ID",
    cls=common.GradientOption,
)
@click.option(
    "--secretAccessKey",
    "secret_access_key",
    help="S3 secret access key",
    cls=common.GradientOption,
)
@click.option(
    "--endpoint",
    "endpoint",
    help="S3 endpoint URL",
    cls=common.GradientOption,
)
@click.option(
    "--region",
    "region",
    help="S3 region",
    cls=common.GradientOption,
)
@click.option(
    "--signatureVersion",
    "signature_version",
    help="S3 signature version (ex: v4)",
    cls=common.GradientOption,
)
@common.api_key_option
@common.options_file
def update_s3_storage_provider(
        id,
        name,
        bucket,
        access_key,
        secret_access_key,
        endpoint,
        region,
        signature_version,
        api_key,
        options_file,
):
    command = commands.UpdateS3StorageProviderCommand(api_key=api_key)
    command.execute(
        storage_provider_id=id,
        name=name,
        bucket=bucket,
        access_key=access_key,
        secret_access_key=secret_access_key,
        endpoint=endpoint,
        region=region,
        signature_version=signature_version,
    )


@storage_providers.command("delete", help="Delete storage provider")
@click.option(
    "--id",
    "id",
    help="Storage provider ID",
    cls=common.GradientOption,
    required=True,
)
@common.api_key_option
@common.options_file
def delete_storage_provider(id, api_key, options_file):
    command = commands.DeleteStorageProviderCommand(api_key=api_key)
    command.execute(id)



import attr


@attr.s
class Machine(object):
    id = attr.ib(type=str, default=None)
    name = attr.ib(type=str, default=None)
    machine_type = attr.ib(type=str, default=None)
    region = attr.ib(type=str, default=None)
    size = attr.ib(type=int, default=None)
    billing_type = attr.ib(type=str, default=None)
    template_id = attr.ib(type=str, default=None)
    assign_public_ip = attr.ib(type=bool, default=None)
    dynamic_public_ip = attr.ib(type=str, default=None)
    network_id = attr.ib(type=str, default=None)
    team_id = attr.ib(type=str, default=None)
    user_id = attr.ib(type=str, default=None)
    email = attr.ib(type=str, default=None)
    password = attr.ib(type=str, default=None)
    first_name = attr.ib(type=str, default=None)
    last_name = attr.ib(type=str, default=None)
    notification_email = attr.ib(type=str, default=None)
    script_id = attr.ib(type=str, default=None)
    os = attr.ib(type=str, default=None)
    cpus = attr.ib(type=str, default=None)
    ram = attr.ib(type=int, default=None)
    gpu = attr.ib(type=str, default=None)
    state = attr.ib(type=str, default=None)
    updates_pending = attr.ib(type=bool, default=None)
    perform_auto_snapshot = attr.ib(type=str, default=None)
    auto_snapshot_frequency = attr.ib(type=str, default=None)
    auto_snapshot_save_count = attr.ib(type=int, default=None)
    shutdown_timeout_in_hours = attr.ib(type=str, default=None)
    shutdown_timeout_forces = attr.ib(type=bool, default=None)
    agent_type = attr.ib(type=str, default=None)
    storage_total = attr.ib(type=str, default=None)
    storage_used = attr.ib(type=str, default=None)
    public_ip_address = attr.ib(type=str, default=None)
    private_ip_address = attr.ib(type=int, default=None)
    usage_rate = attr.ib(type=int, default=None)
    created_timestamp = attr.ib(type=str, default=None)
    last_run_timestamp = attr.ib(type=str, default=None)
    events = attr.ib(type=list, default=None)
    tags = attr.ib(type=list, factory=list)


@attr.s
class MachineEvent(object):
    name = attr.ib(type=str, default=None)
    state = attr.ib(type=str, default=None)
    created = attr.ib(type=str, default=None)


@attr.s
class MachineUtilization(object):
    machine_id = attr.ib(type=str, default=None)
    machine_seconds_used = attr.ib(type=float, default=None)
    machine_billing_month = attr.ib(type=str, default=None)
    machine_hourly_rate = attr.ib(type=str, default=None)
    storage_seconds_used = attr.ib(type=float, default=None)
    storage_billing_month = attr.ib(type=str, default=None)
    storage_monthly_rate = attr.ib(type=str, default=None)



from .artifact import ArtifactSchema
from .dataset import DatasetSchema, DatasetRefSchema
from .dataset_tag import DatasetTagSchema
from .dataset_version import DatasetVersionSchema, DatasetVersionPreSignedURLSchema
from .deployment import DeploymentSchema, DeploymentCreateSchema
from .experiment import BaseExperimentSchema, MultiNodeExperimentSchema, SingleNodeExperimentSchema, \
    MpiMultiNodeExperimentSchema
from .hyperparameter import HyperparameterSchema
from .job import JobSchema
from .log import LogRowSchema
from .machine import MachineSchema, MachineSchemaForListing, MachineEventSchema
from .model import Model, ModelFileSchema
from .notebook import NotebookSchema, NotebookStartSchema
from .project import Project
from .secret import SecretSchema
from .storage_provider import StorageProviderSchema
from .tag import TagSchema
from .tensorboard import InstanceSchema, TensorboardSchema, TensorboardDetailSchema
from .vm_type import VmTypeSchema, VmTypeGpuModelSchema
from .workflows import WorkflowSchema, WorkflowRunSchema, WorkflowSpecSchema



import mock
import pytest

from gradient.commands.experiments import TensorboardHandler


@pytest.fixture()
def tensorboard_handler():
    tensorboard_handler = TensorboardHandler("some_key")
    return tensorboard_handler


class TestTensorboardHandler(object):
    def test_should_add_experiment_to_tensorboard_when_tensorboard_is_a_string(self, tensorboard_handler):
        tensorboard_handler._add_experiment_to_tensorboard = mock.MagicMock()
        tensorboard_handler._create_tensorboard_with_experiment = mock.MagicMock()
        tensorboard_handler._get_tensorboards = mock.MagicMock()

        tensorboard_handler.maybe_add_to_tensorboard("some_tensorboard_id", "some_experiment_id")

        tensorboard_handler._add_experiment_to_tensorboard.assert_called_once_with("some_tensorboard_id",
                                                                                   "some_experiment_id")
        tensorboard_handler._create_tensorboard_with_experiment.assert_not_called()
        tensorboard_handler._get_tensorboards.assert_not_called()

    def test_should_add_experiment_to_tensorboard_when_tensorboard_is_not_a_string_and_one_tb_exists(
            self, tensorboard_handler):
        tensorboard_handler._add_experiment_to_tensorboard = mock.MagicMock()
        tensorboard_handler._create_tensorboard_with_experiment = mock.MagicMock()
        fake_tensorboard = mock.Mock()
        fake_tensorboard.id = "some_tensorboard_id"
        tensorboards = [fake_tensorboard]
        tensorboard_handler._get_tensorboards = mock.MagicMock(return_value=tensorboards)

        tensorboard_handler.maybe_add_to_tensorboard(True, "some_experiment_id")

        tensorboard_handler._get_tensorboards.assert_called_once()
        tensorboard_handler._add_experiment_to_tensorboard.assert_called_once_with("some_tensorboard_id",
                                                                                   "some_experiment_id")
        tensorboard_handler._create_tensorboard_with_experiment.assert_not_called()

    def test_should_add_experiment_to_tensorboard_when_tensorboard_is_not_a_string_and_more_than_one_tb_exists(
            self, tensorboard_handler):
        tensorboard_handler._add_experiment_to_tensorboard = mock.MagicMock()
        tensorboard_handler._create_tensorboard_with_experiment = mock.MagicMock()
        fake_tensorboard = mock.Mock()
        fake_tensorboard.id = "some_tensorboard_id"
        tensorboards = [fake_tensorboard, fake_tensorboard]
        tensorboard_handler._get_tensorboards = mock.MagicMock(return_value=tensorboards)

        tensorboard_handler.maybe_add_to_tensorboard(True, "some_experiment_id")

        tensorboard_handler._get_tensorboards.assert_called_once()
        tensorboard_handler._add_experiment_to_tensorboard.assert_not_called()
        tensorboard_handler._create_tensorboard_with_experiment.assert_called_once_with("some_experiment_id")

    def test_should_add_experiment_to_tensorboard_when_tensorboard_is_not_a_string_and_no_tb_exists(
            self, tensorboard_handler):
        tensorboard_handler._add_experiment_to_tensorboard = mock.MagicMock()
        tensorboard_handler._create_tensorboard_with_experiment = mock.MagicMock()
        tensorboards = []
        tensorboard_handler._get_tensorboards = mock.MagicMock(return_value=tensorboards)

        tensorboard_handler.maybe_add_to_tensorboard(True, "some_experiment_id")

        tensorboard_handler._get_tensorboards.assert_called_once()
        tensorboard_handler._add_experiment_to_tensorboard.assert_not_called()
        tensorboard_handler._create_tensorboard_with_experiment.assert_called_once_with("some_experiment_id")

    @mock.patch("gradient.commands.tensorboards.AddExperimentToTensorboard")
    def test_should_execute_proper_command_when__add_experiment_to_tensorboard_was_executed(
            self, command_class_patched, tensorboard_handler):
        command = mock.MagicMock()
        command_class_patched.return_value = command

        tensorboard_handler._add_experiment_to_tensorboard("tensorboard_id", "experiment_id")

        command_class_patched.assert_called_once_with(api_key="some_key")
        command.execute.assert_called_once_with("tensorboard_id", ["experiment_id"])

    @mock.patch("gradient.commands.experiments.TensorboardClient")
    def test_should_execute_proper_command_when__get_tensorboards_was_executed(
            self, client_class_patched, tensorboard_handler):
        client = mock.MagicMock()
        client_class_patched.return_value = client

        tensorboard_handler._get_tensorboards()

        client_class_patched.assert_called_once_with(api_key="some_key", logger=tensorboard_handler.logger)
        client.list.assert_called_once_with()

    @mock.patch("gradient.commands.tensorboards.CreateTensorboardCommand")
    def test_should_execute_proper_command_when__create_tensorboard_with_experiment(
            self, client_class_patched, tensorboard_handler):
        command = mock.MagicMock()
        client_class_patched.return_value = command

        tensorboard_handler._create_tensorboard_with_experiment("experiment_id")

        client_class_patched.assert_called_once_with(api_key=tensorboard_handler.api_key)
        command.execute.assert_called_once_with(experiments=["experiment_id"])



import attr

from .experiment import BaseExperiment
from .. import constants


@attr.s
class Hyperparameter(BaseExperiment):
    """
    Hyperparameter job class. Inherits from ``BaseExperiment`` class

    :param int experiment_type_id: experiment type, for hyperparameter experiment set to 4
    :param str tuning_command: Tuning command
    :param str worker_container: Worker container
    :param str worker_machine_type: Worker machine type
    :param str worker_command: Worker command
    :param int worker_count: Worker count
    :param bool worker_use_dockerfile: Flag: use dockerfile
    :param str worker_dockerfile_path: path to dockerfile, if not set default is project root directory
    :param str worker_container_user: Worker container user
    :param str worker_registry_username: Worker registry username
    :param str worker_registry_password: Worker registry password

    :param str hyperparameter_server_machine_type: Hyperparameter server  machine type
    :param str hyperparameter_server_container: Hyperparameter server container
    :param str hyperparameter_server_container_user: Hyperparameter server container user
    :param str hyperparameter_server_registry_username: Hyperparameter server registry username
    :param str hyperparameter_server_registry_password: Hyperparameter server registry password
    :param bool is_preemptible: Flag: is preemptible
    :param str trigger_event_id: GradientCI trigger event id
    :param str dockerfile_path: Path to dockerfile
    :param bool use_dockerfile: Flag: use dockerfile
    """
    experiment_type_id = attr.ib(type=int, default=constants.ExperimentType.HYPERPARAMETER_TUNING)
    tuning_command = attr.ib(type=str, default=None)
    worker_container = attr.ib(type=str, default=None)
    worker_machine_type = attr.ib(type=str, default=None)
    worker_command = attr.ib(type=str, default=None)
    worker_count = attr.ib(type=int, default=None)
    worker_use_dockerfile = attr.ib(type=bool, default=None)
    worker_dockerfile_path = attr.ib(type=str, default=None)
    worker_container_user = attr.ib(type=str, default=None)
    worker_registry_username = attr.ib(type=str, default=None)
    worker_registry_password = attr.ib(type=str, default=None)

    hyperparameter_server_machine_type = attr.ib(type=str, default=None)
    hyperparameter_server_container = attr.ib(type=str, default=None)
    hyperparameter_server_container_user = attr.ib(type=str, default=None)
    hyperparameter_server_registry_username = attr.ib(type=str, default=None)
    hyperparameter_server_registry_password = attr.ib(type=str, default=None)

    is_preemptible = attr.ib(type=bool, default=None)
    trigger_event_id = attr.ib(type=str, default=None)
    dockerfile_path = attr.ib(type=str, default=None)
    use_dockerfile = attr.ib(type=bool, default=None)
    tags = attr.ib(type=list, factory=list)

    @experiment_type_id.validator
    def experiment_type_id_validator(self, attribute, value):
        if value is not constants.ExperimentType.HYPERPARAMETER_TUNING:
            raise ValueError("Hyperparameter tuning model's type must equal {}".
                             format(constants.ExperimentType.HYPERPARAMETER_TUNING))



from functools import reduce

import click

from gradient import clilogger
from gradient.api_sdk import constants
from gradient.cli import common
from gradient.cli.cli import cli
from gradient.cli.cli_types import json_string, ChoiceType
from gradient.cli.common import (
    api_key_option, del_if_value_is_none, ClickGroup, jsonify_dicts,
    validate_comma_split_option,
)
from gradient.commands import jobs as jobs_commands
from gradient.commands.jobs import JobAddTagsCommand, JobRemoveTagsCommand, StreamJobMetricsCommand, \
    GetJobMetricsCommand, ListJobMetricsCommand
from gradient.api_sdk.workspace import WorkspaceHandler


def get_workspace_handler():
    logger_ = clilogger.CliLogger()
    workspace_handler = WorkspaceHandler(logger_=logger_)
    return workspace_handler


@cli.group("jobs", help="Manage gradient jobs", cls=ClickGroup)
def jobs_group():
    pass


@jobs_group.group("tags", help="Manage job tags", cls=ClickGroup)
def jobs_tags():
    pass


@jobs_group.group(name="metrics", help="Read job metrics", cls=ClickGroup)
def jobs_metrics():
    pass


@jobs_group.command("delete", help="Delete job")
@click.option(
    "--id",
    "job_id",
    required=True,
    help="Delete job with given ID",
)
@api_key_option
def delete_job(job_id, api_key):
    command = jobs_commands.DeleteJobCommand(api_key=api_key)
    command.execute(job_id)


@jobs_group.command("stop", help="Stop running job")
@click.option(
    "--id",
    "job_id",
    required=True,
    help="Stop job with given ID",
)
@api_key_option
def stop_job(job_id, api_key=None):
    command = jobs_commands.StopJobCommand(api_key=api_key)
    command.execute(job_id)


@jobs_group.command("list", help="List jobs with optional filtering")
@click.option(
    "--project",
    "project",
    help="Use to filter jobs by project name",
    cls=common.GradientOption,
)
@click.option(
    "--projectId",
    "project_id",
    help="Use to filter jobs by project ID",
    cls=common.GradientOption,
)
@click.option(
    "--experimentId",
    "experiment_id",
    help="Use to filter jobs by experiment ID",
    cls=common.GradientOption,
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    cls=common.GradientOption,
    help="Filter by tags. Multiple use"
)
@api_key_option
@common.options_file
def list_jobs(api_key, options_file, **filters):
    del_if_value_is_none(filters)

    command = jobs_commands.ListJobsCommand(api_key=api_key)
    command.execute(**filters)


def common_jobs_create_options(f):
    options = [
        click.option(
            "--name",
            "name",
            help="Job name",
            cls=common.GradientOption,
        ),
        click.option(
            "--machineType",
            "machine_type",
            help="Virtual machine type",
            required=True,
            cls=common.GradientOption,
        ),
        click.option(
            "--container",
            "container",
            default="paperspace/tensorflow-python",
            help="Docker container",
            required=True,
            cls=common.GradientOption,
        ),
        click.option(
            "--command",
            "command",
            help="Job command/entrypoint",
            cls=common.GradientOption,
        ),
        click.option(
            "--ports",
            "ports",
            help="Mapped ports",
            cls=common.GradientOption,
        ),
        # TODO: make it a flag
        click.option(
            "--isPublic",
            "is_public",
            help="Flag: is job public",
            cls=common.GradientOption,
        ),
        click.option(
            "--workspace",
            "workspace",
            help="Path to workspace directory",
            cls=common.GradientOption,
        ),
        click.option(
            "--workingDirectory",
            "working_directory",
            help="Working directory for the experiment",
            cls=common.GradientOption,
        ),
        click.option(
            "--ignoreFiles",
            "ignore_files",
            help="Ignore certain files from uploading",
            cls=common.GradientOption,
        ),
        click.option(
            "--experimentId",
            "experiment_id",
            help="Experiment Id",
            cls=common.GradientOption,
        ),
        click.option(
            "--envVars",
            "env_vars",
            type=json_string,
            help="Environmental variables",
            cls=common.GradientOption,
        ),
        click.option(
            "--useDockerfile",
            "use_dockerfile",
            help="Flag: using Dockerfile",
            cls=common.GradientOption,
        ),
        # TODO: make it a flag
        click.option(
            "--isPreemptible",
            "is_preemptible",
            help="Flag: isPreemptible",
            cls=common.GradientOption,
        ),
        click.option(
            "--project",
            "project",
            help="Project name",
            cls=common.GradientOption,
        ),
        click.option(
            "--projectId",
            "project_id",
            help="Project ID",
            required=True,
            cls=common.GradientOption,
        ),
        click.option(
            "--startedByUserId",
            "started_by_user_id",
            help="User ID",
            cls=common.GradientOption,
        ),
        click.option(
            "--relDockerfilePath",
            "rel_dockerfile_path",
            help="Relative path to Dockerfile",
            cls=common.GradientOption,
        ),
        click.option(
            "--registryUsername",
            "registry_username",
            help="Docker registry username",
            cls=common.GradientOption,
        ),
        click.option(
            "--registryPassword",
            "registry_password",
            help="Docker registry password",
            cls=common.GradientOption,
        ),
        click.option(
            "--cluster",
            "cluster",
            help="Cluster name",
            cls=common.GradientOption,
        ),
        click.option(
            "--clusterId",
            "cluster_id",
            help="Cluster id",
            cls=common.GradientOption,
        ),
        click.option(
            "--nodeAttrs",
            "node_attrs",
            type=json_string,
            help="Cluster node details",
            cls=common.GradientOption,
        ),
        click.option(
            "--registryTarget",
            "registry_target",
            help="Docker registry target",
            cls=common.GradientOption,
        ),
        click.option(
            "--registryTargetUsername",
            "registry_target_username",
            help="Docker registry username",
            cls=common.GradientOption,
        ),
        click.option(
            "--registryTargetPassword",
            "registry_target_password",
            help="Docker registry password",
            cls=common.GradientOption,
        ),
        click.option(
            "--buildOnly",
            "build_only",
            type=bool,
            is_flag=True,
            help="Determines whether to only build and not run image (default false)",
            cls=common.GradientOption,
        ),
        click.option(
            "--tag",
            "tags",
            multiple=True,
            help="One or many tags that you want to add to experiment",
            cls=common.GradientOption
        ),
        click.option(
            "--tags",
            "tags_comma",
            help="Separated by comma tags that you want add to experiment",
            cls=common.GradientOption
        ),
        click.option(
            "--dataset",
            "datasets",
            help="Separated by comma tags that you want add to experiment",
            cls=common.GradientOption,
            multiple=True,
        )
    ]
    return reduce(lambda x, opt: opt(x), reversed(options), f)


@jobs_group.command("create", help="Create job")
@common_jobs_create_options
@api_key_option
@common.options_file
@click.pass_context
def create_job(ctx, api_key, options_file, datasets=None, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"))

    del_if_value_is_none(kwargs)
    jsonify_dicts(kwargs)

    if datasets:
        values = []

        for value in datasets:
            name, _, ref = value.partition("@")
            if not (name and ref):
                raise click.UsageError(
                    "Dataset '%s' must have an @ (ex: images@dsr8k5qzn401lb5:klfoyy9)" % value)

            dataset = {"id": ref, "name": name}
            if ":" not in ref:
                dataset["output"] = True

            values.append(dataset)

        kwargs["datasets"] = values



    command = jobs_commands.CreateJobCommand(api_key=api_key, workspace_handler=get_workspace_handler())
    job_handle = command.execute(kwargs)
    if job_handle is not None:
        ctx.invoke(list_logs, job_id=job_handle, line=1, limit=100, follow=True, api_key=api_key)


@jobs_group.command("logs", help="List job logs")
@click.option(
    "--id",
    "job_id",
    required=True,
    cls=common.GradientOption,
)
@click.option(
    "--line",
    "line",
    required=False,
    default=0,
    cls=common.GradientOption,
)
@click.option(
    "--limit",
    "limit",
    required=False,
    default=10000,
    cls=common.GradientOption,
)
@click.option(
    "--follow",
    "follow",
    required=False,
    default=False,
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def list_logs(job_id, line, limit, follow, options_file, api_key=None):
    command = jobs_commands.JobLogsCommand(api_key=api_key)
    command.execute(job_id, line, limit, follow)


@jobs_group.group("artifacts", help="Manage jobs' artifacts", cls=ClickGroup)
def artifacts():
    pass


@artifacts.command("destroy", help="Destroy job's artifacts")
@click.option(
    "--id",
    "job_id",
    cls=common.GradientOption,
    help="ID of the job",
)
@click.option(
    "--files",
    "files",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def destroy_artifacts(job_id, options_file, api_key=None, files=None):
    command = jobs_commands.ArtifactsDestroyCommand(api_key=api_key)
    command.execute(job_id, files=files)


@artifacts.command("get", help="Get job's artifacts")
@click.option(
    "--id",
    "job_id",
    cls=common.GradientOption,
    help="ID of the job",
)
@api_key_option
@common.options_file
def get_artifacts(job_id, options_file, api_key=None):
    command = jobs_commands.ArtifactsGetCommand(api_key=api_key)
    command.execute(job_id)


@artifacts.command("list", help="List job's artifacts")
@click.option(
    "--id",
    "job_id",
    cls=common.GradientOption,
    help="ID of the job",
)
@click.option(
    "--size",
    "-s",
    "size",
    help="Show file size",
    is_flag=True,
    cls=common.GradientOption,
)
@click.option(
    "--links",
    "-l",
    "links",
    help="Show file URL",
    is_flag=True,
    default=False,
    cls=common.GradientOption,
)
@click.option(
    "--files",
    "files",
    help="Get only given file (use at the end * as a wildcard)",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def list_artifacts(job_id, size, links, files, options_file, api_key=None):
    command = jobs_commands.ArtifactsListCommand(api_key=api_key)
    command.execute(job_id=job_id, size=size, links=links, files=files)


@artifacts.command("download", help="List job's artifacts")
@click.option(
    "--id",
    "job_id",
    cls=common.GradientOption,
)
@click.option(
    "--destinationDir",
    "destination_directory",
    required=True,
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def download_artifacts(job_id, destination_directory, options_file, api_key=None):
    command = jobs_commands.DownloadArtifactsCommand(api_key=api_key)
    command.execute(job_id=job_id, destination_directory=destination_directory)


@jobs_tags.command("add", help="Add tags to job")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the job",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to add to job",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want add to job",
    cls=common.GradientOption
)
@api_key_option
@common.options_file
def job_add_tag(id, options_file, api_key, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"), raise_if_no_values=True)

    command = JobAddTagsCommand(api_key=api_key)
    command.execute(id, **kwargs)


@jobs_tags.command("remove", help="Remove tags from job")
@click.option(
    "--id",
    "id",
    required=True,
    cls=common.GradientOption,
    help="ID of the job",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="One or many tags that you want to remove from job",
    cls=common.GradientOption
)
@click.option(
    "--tags",
    "tags_comma",
    help="Separated by comma tags that you want to remove from job",
    cls=common.GradientOption
)
@api_key_option
@common.options_file
def job_remove_tags(id, options_file, api_key, **kwargs):
    kwargs["tags"] = validate_comma_split_option(kwargs.pop("tags_comma"), kwargs.pop("tags"), raise_if_no_values=True)

    command = JobRemoveTagsCommand(api_key=api_key)
    command.execute(id, **kwargs)



@jobs_metrics.command(
    "get",
    short_help="Get job metrics",
    help="Get job metrics. Shows CPU and RAM usage by default",
)
@click.option(
    "--id",
    "job_id",
    required=True,
    cls=common.GradientOption,
    help="ID of the job",
)
@click.option(
    "--metric",
    "metrics_list",
    multiple=True,
    type=str,
    default=(constants.BuiltinMetrics.cpu_percentage, constants.BuiltinMetrics.memory_usage),
    help=("One or more metrics that you want to read: {}. Defaults to cpuPercentage and memoryUsage. To view available custom metrics, use command: `gradient jobs metrics list`".format(', '.join(map(str, constants.METRICS_MAP)))),
    cls=common.GradientOption,
)
@click.option(
    "--interval",
    "interval",
    default="30s",
    help="Interval",
    cls=common.GradientOption,
)
@click.option(
    "--start",
    "start",
    type=click.DateTime(),
    help="Timestamp of first time series metric to collect",
    cls=common.GradientOption,
)
@click.option(
    "--end",
    "end",
    type=click.DateTime(),
    help="Timestamp of last time series metric to collect",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def get_job_metrics(job_id, metrics_list, interval, start, end, options_file, api_key):
    command = GetJobMetricsCommand(api_key=api_key)
    command.execute(job_id, start, end, interval, built_in_metrics=metrics_list)

@jobs_metrics.command(
    "list",
    short_help="List job metrics",
    help="List job metrics. Shows CPU and RAM usage by default",
)
@click.option(
    "--id",
    "job_id",
    required=True,
    cls=common.GradientOption,
    help="ID of the job",
)
@click.option(
    "--interval",
    "interval",
    default="30s",
    help="Interval",
    cls=common.GradientOption,
)
@click.option(
    "--start",
    "start",
    type=click.DateTime(),
    help="Timestamp of first time series metric to collect",
    cls=common.GradientOption,
)
@click.option(
    "--end",
    "end",
    type=click.DateTime(),
    help="Timestamp of last time series metric to collect",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def list_job_metrics(job_id, interval, start, end, options_file, api_key):
    command = ListJobMetricsCommand(api_key=api_key)
    command.execute(job_id, start, end, interval)

@jobs_metrics.command(
    "stream",
    short_help="Watch live job metrics",
    help="Watch live job metrics. Shows CPU and RAM usage by default",
)
@click.option(
    "--id",
    "job_id",
    required=True,
    cls=common.GradientOption,
    help="ID of the job",
)
@click.option(
    "--metric",
    "metrics_list",
    multiple=True,
    type=ChoiceType(constants.METRICS_MAP, case_sensitive=False),
    default=(constants.BuiltinMetrics.cpu_percentage, constants.BuiltinMetrics.memory_usage),
    help="One or more metrics that you want to read. Defaults to cpuPercentage and memoryUsage",
    cls=common.GradientOption,
)
@click.option(
    "--interval",
    "interval",
    default="30s",
    help="Interval",
    cls=common.GradientOption,
)
@api_key_option
@common.options_file
def stream_job_metrics(job_id, metrics_list, interval, options_file, api_key):
    command = StreamJobMetricsCommand(api_key=api_key)
    command.execute(job_id=job_id, interval=interval, built_in_metrics=metrics_list)



from .base_client import BaseClient
from .. import models, repositories


class DatasetsClient(BaseClient):

    def list(self, limit=20, offset=0):
        """Get a list of datasets

        :param int limit: Limit results
        :param int offset: Skip results

        :returns: List of datasets
        :rtype: list[models.Dataset]
        """

        repository = self.build_repository(repositories.ListDatasets)
        return repository.list(limit=limit, offset=offset)

    def delete(self, dataset_id):
        """Delete a dataset

        :param str dataset_id: Dataset ID [required]

        :returns:
        :rtype: None
        """

        repository = self.build_repository(repositories.DeleteDataset)
        repository.delete(dataset_id)

    def get(self, dataset_id):
        """Get a dataset

        :param str dataset_id: Dataset ID [required]

        :returns: Dataset
        :rtype: models.Dataset
        """
        repository = self.build_repository(repositories.GetDataset)
        return repository.get(id=dataset_id)

    def get_ref(self, dataset_ref):
        """Get dataset with resolved version by reference

        :param str dataset_ref: Dataset reference [required]

        :returns: Dataset with resolved version
        :rtype: models.DatasetRef
        """
        repository = self.build_repository(repositories.GetDatasetRef)
        return repository.get(id=dataset_ref)

    def create(self, name, storage_provider_id, description=None):
        """Create a new dataset

        :param str name: Name of dataset [required]
        :param str storage_provider_id: Storage provider ID [required]
        :param str description: Brief description of the dataset

        :returns: Dataset ID
        :rtype: str
        """

        dataset = models.Dataset(
            name=name,
            storage_provider_id=storage_provider_id,
            description=description,
        )

        repository = self.build_repository(repositories.CreateDataset)
        return repository.create(dataset)

    def update(self, dataset_id, name=None, description=None):
        """Update an existing dataset

        :param str dataset_id: Dataset ID [required]
        :param str name: Name of dataset
        :param str description: Brief description of the dataset

        :returns:
        :rtype: None
        """

        dataset = models.Dataset(
            name=name,
            description=description
        )

        repository = self.build_repository(repositories.UpdateDataset)
        return repository.update(dataset_id, dataset)



import marshmallow as ma

from .base import BaseSchema
from .. import models


class WorkflowSchema(BaseSchema):
    MODEL = models.Workflow

    id = ma.fields.Str(dump_to="id", load_from="id")
    team_id = ma.fields.Str(dump_to="teamId", load_from="teamId")
    project_id = ma.fields.Int(required=True, dump_to="projectId", load_from="projectId")
    name = ma.fields.Str(dump_to="name", load_from="name")
    workflow_spec_id = ma.fields.Str(dump_to="workflowSpecId", load_from="workflowSpecId")
    dt_deleted = ma.fields.DateTime(dump_to="dtDeleted", load_from="dtDeleted")
    dt_created = ma.fields.DateTime(dump_to="dtCreated", load_from="dtCreated")
    dt_modified = ma.fields.DateTime(dump_to="dtModified", load_from="dtModified")


class WorkflowSpecSchema(BaseSchema):
    MODEL = models.WorkflowSpec

    id = ma.fields.Str(dump_to="id", load_from="id")
    data = ma.fields.Str(dump_to="data", load_from="data")
    hash_sha256 = ma.fields.Str(dump_to="hashSha256", load_from="hashSha256")
    dt_created = ma.fields.DateTime(dump_to="dtCreated", load_from="dtCreated")


class WorkflowRunSchema(BaseSchema):
    MODEL = models.WorkflowRun

    id = ma.fields.Str(dump_to="id", load_from="id")
    team_id = ma.fields.Int(dump_to="teamId", load_from="teamId")
    workflow_id = ma.fields.Str(dump_to="workflowId", load_from="workflowId")
    cluster_id = ma.fields.Int(required=True, dump_to="clusterId", load_from="clusterId")
    user_id = ma.fields.Int(required=True, dump_to="userId", load_from="userId")
    workflow_spec_id = ma.fields.Str(dump_to="workflowSpecId", load_from="workflowSpecId")
    seq_num = ma.fields.Int(required=True, dump_to="seqNum", load_from="seqNum")
    timeout = ma.fields.Int(required=True, dump_to="timeout", load_from="timeout")
    workflow_phase_id = ma.fields.Int(required=True, dump_to="workflowPhaseId", load_from="workflowPhaseId")
    name = ma.fields.Str(dump_to="name", load_from="name")
    message = ma.fields.Str(dump_to="message", load_from="message")
    dt_status = ma.fields.DateTime(dump_to="dtStatus", load_from="dtStatus")
    dt_started = ma.fields.DateTime(dump_to="dtStarted", load_from="dtStarted")
    dt_finished = ma.fields.DateTime(dump_to="dtFinished", load_from="dtFinished")
    dt_created = ma.fields.DateTime(dump_to="dtCreated", load_from="dtCreated")
    dt_modified = ma.fields.DateTime(dump_to="dtModified", load_from="dtDeleted")



import json
from collections import OrderedDict

from ..config import config
from ..repositories.common import GetResource, ListResources
from ..sdk_exceptions import MalformedResponseError
from ..serializers.cluster import ClusterSchema


class ValidateClusterRepository(GetResource):
    SERIALIZER_CLS = ClusterSchema

    def get_request_url(self, **kwargs):
        return "/clusters/getCluster"

    def _get_api_url(self, **kwargs):
        return config.CONFIG_HOST

    def _get_request_params(self, kwargs):
        return {
            "id": kwargs.get("cluster_id")
        }

    def _parse_object(self, instance_dict, **kwargs):
        """
        :param dict instance_dict:
        :return: model instance
        """
        instance = self.SERIALIZER_CLS().dump(instance_dict)
        if instance.errors:
            raise MalformedResponseError(instance.errors)
        return instance.data


class ListClusters(ListResources):
    SERIALIZER_CLS = ClusterSchema

    def get_request_url(self, **kwargs):
        return "/clusters/getClusters"

    def _get_api_url(self, **kwargs):
        return config.CONFIG_HOST

    def _get_request_params(self, kwargs):
        filter = OrderedDict()
        filter["limit"] = kwargs.get("limit")
        filter["offset"] = kwargs.get("offset")
        filter["where"] = {
            "isPrivate": True
        }

        return {
            "filter": json.dumps(OrderedDict(filter))
        }



import copy

import requests

from gradient import version
from .. import utils, logger as sdk_logger
from ..config import config

default_headers = {"X-API-Key": config.PAPERSPACE_API_KEY,
                   "ps_client_name": "gradient-cli-sdk",
                   "ps_client_version": version.version}


class API(object):
    def __init__(self, api_url, headers=None, api_key=None, ps_client_name=None, logger=sdk_logger.MuteLogger()):
        """

        :param str api_url: url you want to connect
        :param dict headers: headers
        :param str api_key: your API key
        :param str ps_client_name: Client name
        :param sdk_logger.Logger logger:
        """
        self.api_url = api_url
        headers = headers or default_headers
        self.headers = headers.copy()

        if api_key:
            self.api_key = api_key

        if ps_client_name is not None:
            self.ps_client_name = ps_client_name

        self.logger = logger

    @property
    def api_key(self):
        return self.headers.get("X-API-Key")

    @api_key.setter
    def api_key(self, value):
        self.headers["X-API-Key"] = value

    @property
    def ps_client_name(self):
        return self.headers.get("ps_client_name")

    @ps_client_name.setter
    def ps_client_name(self, value):
        self.headers["ps_client_name"] = value

    def get_path(self, url):
        if not url:
            return self.api_url

        full_path = utils.concatenate_urls(self.api_url, url)
        return full_path

    def post(self, url, json=None, params=None, files=None, data=None):
        path = self.get_path(url)
        headers = copy.deepcopy(self.headers)
        if data:
            headers["Content-Type"] = data.content_type

        self.logger.debug("POST request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}\n\tfiles: {}\n\tdata: {}"
                          .format(path, headers, json, params, files, data))
        response = requests.post(path, json=json, params=params, headers=headers, files=files, data=data)
        self.logger.debug("Response status code: {}".format(response.status_code))
        self.logger.debug("Response content: {}".format(response.content))
        return response

    def put(self, url, json=None, params=None, data=None):
        path = self.get_path(url)
        self.logger.debug("PUT request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}"
                          .format(path, self.headers, json, params))
        response = requests.put(path, json=json, params=params, headers=self.headers, data=data)
        self.logger.debug("Response status code: {}".format(response.status_code))
        self.logger.debug("Response content: {}".format(response.content))
        return response

    def get(self, url, json=None, params=None):
        path = self.get_path(url)
        self.logger.debug("GET request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}"
                          .format(path, self.headers, json, params))
        response = requests.get(path, params=params, headers=self.headers, json=json)
        self.logger.debug("Response status code: {}".format(response.status_code))
        self.logger.debug("Response content: {}".format(response.content))
        return response

    def delete(self, url, json=None, params=None):
        path = self.get_path(url)
        response = requests.delete(path, params=params, headers=self.headers, json=json)
        self.logger.debug("DELETE request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}"
                          .format(response.url, self.headers, json, params))
        self.logger.debug("Response status code: {}".format(response.status_code))
        self.logger.debug("Response content: {}".format(response.content))
        return response


class GradientResponse(object):
    def __init__(self, body, code, headers, data, request=None):
        self.body = body
        self.code = code
        self.headers = headers
        self.data = data
        self.request = request

    @property
    def ok(self):
        return 200 <= self.code < 400

    @classmethod
    def interpret_response(cls, response):
        """
        :type response: requests.Response
        :rtype: GradientResponse
        """
        try:
            data = response.json()
        except ValueError:
            content = response.content
            data = content or None

        gradient_response = cls(response.content, response.status_code, response.headers, data,
                                request=response.request)
        return gradient_response



from .common import BaseRepository, ListResources, GetResource, ListLogs
from .. import config, serializers
from ..clients import http_client
import json

class WorkflowsMixin(object):
    SERIALIZER_CLS = serializers.WorkflowSchema

    @staticmethod
    def _get_api_url(**kwargs):
        return config.config.CONFIG_HOST


class ListWorkflows(WorkflowsMixin, ListResources):
    def get_request_url(self, **kwargs):
        project_id = kwargs.get("project_id")
        if project_id is not None:
            return "/workflows?filter[where][projectId]={}".format(project_id)
        return "/workflows"

    def _get_instances(self, response, **kwargs):
        if not response.data:
            return []

        objects = self._parse_objects(response.data, **kwargs)
        return objects

class GetWorkflow(WorkflowsMixin, BaseRepository):
    def get_request_url(self, **kwargs):
        return "/workflows/{}".format(kwargs.get("id"))

    def get(self, **kwargs):
        json_ = self._get_request_json(kwargs)
        params = self._get_request_params(kwargs)
        url = self.get_request_url(**kwargs)
        client = self._get_client(**kwargs)
        response = self._send_request(client, url, json=json_, params=params)
        gradient_response = http_client.GradientResponse.interpret_response(response)

        if not gradient_response.data:
            return {}

        return gradient_response.data


class WorkflowRunsMixin(object):
    @staticmethod
    def _get_api_url(**kwargs):
        return config.config.CONFIG_HOST

class ListWorkflowRuns(WorkflowRunsMixin, BaseRepository):
    @staticmethod
    def _get_api_url(**kwargs):
        return config.config.CONFIG_HOST


    def get_request_url(self, **kwargs):
        return "/workflows/{}/runs".format(kwargs.get("id"))

    def get(self, **kwargs):
        json_ = self._get_request_json(kwargs)
        params = self._get_request_params(kwargs)
        url = self.get_request_url(**kwargs)
        client = self._get_client(**kwargs)
        response = self._send_request(client, url, json=json_, params=params)
        gradient_response = http_client.GradientResponse.interpret_response(response)

        if not gradient_response.data:
            return []

        return gradient_response.data

class GetWorkflowRun(WorkflowRunsMixin, BaseRepository):
    def get_request_url(self, **kwargs):
        return "/workflows/{}/runs/{}".format(kwargs.get("id"), kwargs.get("run"))

    def get(self, **kwargs):
        json_ = self._get_request_json(kwargs)
        params = self._get_request_params(kwargs)
        url = self.get_request_url(**kwargs)
        client = self._get_client(**kwargs)
        response = self._send_request(client, url, json=json_, params=params)
        gradient_response = http_client.GradientResponse.interpret_response(response)

        if not gradient_response.data:
            return {}

        return gradient_response.data


class CreateWorkflow(WorkflowsMixin, BaseRepository):
    def get_request_url(self, **kwargs):
        return "/workflows"

    def _get_request_json(self, kwargs):
        return {"name": kwargs.get("name"), "projectId": kwargs.get("project_id")}

    def _send_request(self, client, url, json=None, params=None):
        response = client.post(url, json=json, params=params)
        return response

    def create(self, **kwargs):
        response = self._get(**kwargs)
        self._validate_response(response)

        if not response.data:
            return {}

        return response.data
        
class CreateWorkflowRun(WorkflowsMixin, BaseRepository):
    def get_request_url(self, **kwargs):
        return "/workflows/{}/runs".format(kwargs.get("id"))

    def _get_request_json(self, **kwargs):
        if kwargs.get("inputs") is not None:
            return {"spec": kwargs.get("spec"), "clusterId": kwargs.get("cluster_id"), "run": True, "markDefault": False, "inputs": kwargs.get("inputs") }

        return {"spec": kwargs.get("spec"), "clusterId": kwargs.get("cluster_id"), "run": True, "markDefault": False }

    def _send_create_request(self, **kwargs):
        url = self.get_request_url(**kwargs)
        client = self._get_client(**kwargs)
        json_ = self._get_request_json(**kwargs)

        response = client.post(url, json=json_)
        gradient_response = http_client.GradientResponse.interpret_response(response)
        json_formatted_str = json.dumps(gradient_response.data, indent=4)
        return gradient_response

    def create(self, **kwargs):
        response = self._send_create_request(**kwargs)
        self._validate_response(response)

        if not response.data:
            return {}

        return response.data

class ListWorkflowLogs(ListLogs):
    def _get_request_params(self, kwargs):
        params = {
            "jobId": kwargs["id"],
            "line": kwargs["line"],
            "limit": kwargs["limit"],
        }
        return params


import marshmallow
from enum import Enum
from marshmallow.validate import OneOf

from .base import BaseSchema
from .tag import TagSchema
from .. import models


class InstanceType(Enum):
    CPU = 'cpu'
    GPU = 'gpu'


class InstanceSize(Enum):
    SMALL = 'small'
    MEDIUM = 'medium'
    LARGE = 'large'


class InstanceSchema(BaseSchema):
    MODEL = models.Instance

    type = marshmallow.fields.String(validate=marshmallow.validate.OneOf(choices=[e.value for e in InstanceType]))
    size = marshmallow.fields.String(validate=marshmallow.validate.OneOf(choices=[e.value for e in InstanceSize]))
    count = marshmallow.fields.Int()


class TBExperimentSchema(BaseSchema):
    id = marshmallow.fields.String()
    project_id = marshmallow.fields.String()
    state = marshmallow.fields.String()


class TensorboardDetailSchema(BaseSchema):
    MODEL = models.Tensorboard

    id = marshmallow.fields.Str()
    image = marshmallow.fields.Str()
    username = marshmallow.fields.Str()
    password = marshmallow.fields.Str()
    instance = marshmallow.fields.Nested(InstanceSchema, required=False, default=None)
    experiments = marshmallow.fields.List(marshmallow.fields.Nested(TBExperimentSchema))
    url = marshmallow.fields.Str()
    state = marshmallow.fields.String()
    tags = marshmallow.fields.Nested(TagSchema, only="name", many=True, load_only=True)


class TensorboardSchema(BaseSchema):
    MODEL = models.Tensorboard

    id = marshmallow.fields.Str()
    image = marshmallow.fields.Str()
    username = marshmallow.fields.Str()
    password = marshmallow.fields.Str()
    instance = marshmallow.fields.Nested(InstanceSchema, required=False, default=None)
    experiments = marshmallow.fields.List(marshmallow.fields.String())
    url = marshmallow.fields.Str()
    state = marshmallow.fields.String()



from .common import BaseRepository, AlterResource, CreateResource, DeleteResource, GetResource, ListResources
from .datasets import DatasetMixin
from .. import serializers


class DatasetVersionMixin(DatasetMixin):
    SERIALIZER_CLS = serializers.DatasetVersionSchema

    @classmethod
    def get_request_url(cls, id="", **kwargs):
        dataset_id, _, version = id.partition(":")

        url = super(DatasetVersionMixin, cls).get_request_url(id=dataset_id, **kwargs) + "/versions"
        if version:
            url += "/{}".format(version)
        return url


class ListDatasetVersions(DatasetVersionMixin, ListResources):
    def _get_request_params(self, kwargs):
        limit = kwargs.get("limit") or 20
        offset = kwargs.get("offset") or 0
        is_committed = bool(kwargs.get("is_committed", True))
        return {
            "filter[where][isCommitted]": str(is_committed).lower(),
            "filter[limit]": limit,
            "filter[skip]": offset,
            "filter[order][]": "dtCreated DESC",
        }


class CreateDatasetVersion(DatasetVersionMixin, CreateResource):
    HANDLE_FIELD = "version"

    def _get_request_json(self, instance_dict):
        instance_dict = dict(instance_dict)
        instance_dict.pop("id", None)
        return instance_dict


class GetDatasetVersion(DatasetVersionMixin, GetResource):
    pass


class UpdateDatasetVersion(DatasetVersionMixin, AlterResource):
    def _get_request_json(self, kwargs):
        data = dict(kwargs)
        if "id" in data:
            del data["id"]
        return data


class DeleteDatasetVersion(DatasetVersionMixin, DeleteResource):
    pass


class GenerateDatasetVersionPreSignedS3Urls(DatasetVersionMixin, BaseRepository):
    @classmethod
    def get_request_url(cls, id=None, **kwargs):
        return super(GenerateDatasetVersionPreSignedS3Urls, cls).get_request_url(id=id) + "/s3/preSignedUrls"

    def generate(self, id, calls):
        response = self._get(id=id, calls=calls)
        self._validate_response(response)
        return serializers.DatasetVersionPreSignedURLSchema().get_instance(response.data, many=True)

    def _get_request_json(self, kwargs):
        return {'calls': kwargs['calls']}

    def _send_request(self, client, url, json=None, params=None):
        return client.post(url, json=json, params=params)



from . import DeploymentsClient, ExperimentsClient, HyperparameterJobsClient, ModelsClient, ProjectsClient, \
    MachinesClient, NotebooksClient, SecretsClient
from .job_client import JobsClient
from .workflow_client import WorkflowsClient
from .. import logger as sdk_logger


class SdkClient(object):
    def __init__(self, api_key, logger=sdk_logger.MuteLogger()):
        """
        :param str api_key: API key
        :param sdk_logger.Logger logger:
        """
        self.experiments = ExperimentsClient(api_key=api_key, logger=logger)
        self.deployments = DeploymentsClient(api_key=api_key, logger=logger)
        self.hyperparameters = HyperparameterJobsClient(api_key=api_key, logger=logger)
        self.models = ModelsClient(api_key=api_key, logger=logger)
        self.jobs = JobsClient(api_key=api_key, logger=logger)
        self.projects = ProjectsClient(api_key=api_key, logger=logger)
        self.machines = MachinesClient(api_key=api_key, logger=logger)
        self.notebooks = NotebooksClient(api_key=api_key, logger=logger)
        self.secrets = SecretsClient(api_key=api_key, logger=logger)
        self.workflows = WorkflowsClient(api_key=api_key, logger=logger)



import mock
from click.testing import CliRunner

from gradient.api_sdk.clients import http_client
from gradient.cli import cli
from tests import example_responses, MockResponse

EXPECTED_HEADERS = http_client.default_headers.copy()
EXPECTED_HEADERS["ps_client_name"] = "gradient-cli"

EXPECTED_HEADERS_WITH_CHANGED_API_KEY = EXPECTED_HEADERS.copy()
EXPECTED_HEADERS_WITH_CHANGED_API_KEY["X-API-Key"] = "some_key"

URL = "https://api.paperspace.io"


class TestSecretsList(object):
    COMMAND = ["secrets", "list"]

    LIST_SECRETS = example_responses.LIST_SECRETS_RESPONSE

    LIST_STDOUT = """+-----------------------+
| Name                  |
+-----------------------+
| aws_access_key_id     |
| aws_secret_access_key |
+-----------------------+
"""

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_list_of_secrets(self, get_patched):
        get_patched.return_value = MockResponse(self.LIST_SECRETS)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND + ["team"])

        assert self.LIST_STDOUT in result.output, result.exc_info
        get_patched.assert_called_once_with(URL + "/teams/secrets",
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params={})

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_list_of_secrets_with_id(self, get_patched):
        get_patched.return_value = MockResponse(self.LIST_SECRETS)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND + ["team", "--id", "te1234567"])

        assert self.LIST_STDOUT in result.output, result.exc_info
        get_patched.assert_called_once_with(URL + "/teams/secrets",
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params={"teamId": "te1234567"})


class TestSecretsSet(object):
    COMMAND = ["secrets", "set"]

    SET_STDOUT = "Set {} secret 'aws_access_key_id'\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    def test_should_send_put_request_and_print_status(self, put_patched):
        put_patched.return_value = MockResponse()

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND + ["team", "--name=aws_access_key_id", "--value=secret"])

        assert self.SET_STDOUT.format("team") in result.output, result.exc_info
        put_patched.assert_called_once_with(URL + "/teams/secrets/aws_access_key_id",
                                            headers=EXPECTED_HEADERS,
                                            params={},
                                            json={"value": "secret"},
                                            data=None)


    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    def test_should_send_put_request_and_print_status_with_id(self, put_patched):
        put_patched.return_value = MockResponse()

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND + ["project", "--id=pr1234567",
                                                        "--name=aws_access_key_id", "--value=secret"])

        assert self.SET_STDOUT.format("project") in result.output, result.exc_info
        put_patched.assert_called_once_with(URL + "/projects/secrets/aws_access_key_id",
                                            headers=EXPECTED_HEADERS,
                                            params={"projectId": "pr1234567"},
                                            json={"value": "secret"},
                                            data=None)


class TestSecretsDelete(object):
    COMMAND = ["secrets", "delete"]

    SET_STDOUT = "Deleted cluster secret 'aws_secret_access_key'\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.delete")
    def test_should_send_delete_request_and_print_status_with_id(self, delete_patched):
        delete_patched.return_value = MockResponse()

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND + ["team", "--name=aws_secret_access_key"])

        assert self.SET_STDOUT in result.output, result.exc_info
        delete_patched.assert_called_once_with(URL + "/teams/secrets/aws_secret_access_key",
                                               headers=EXPECTED_HEADERS,
                                               params={},
                                               json=None)

    @mock.patch("gradient.api_sdk.clients.http_client.requests.delete")
    def test_should_send_delete_request_and_print_status_with_id(self, delete_patched):
        delete_patched.return_value = MockResponse()

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND + ["cluster", "--id=cl1234567",
                                                        "--name=aws_secret_access_key"])

        assert self.SET_STDOUT in result.output, result.exc_info
        delete_patched.assert_called_once_with(URL + "/clusters/secrets/aws_secret_access_key",
                                               headers=EXPECTED_HEADERS,
                                               params={"clusterId": "cl1234567"},
                                               json=None)



import attr


@attr.s
class JobDataset(object):
    id = attr.ib(type=str, default=None)
    name = attr.ib(type=str, default=None)
    output = attr.ib(type=bool, default=None)


@attr.s
class Job(object):
    id = attr.ib(type=str, default=None)
    name = attr.ib(type=str, default=None)
    state = attr.ib(type=str, default=None)
    working_directory = attr.ib(type=str, default=None)
    artifacts_directory = attr.ib(type=str, default=None)
    entrypoint = attr.ib(type=str, default=None)
    project_id = attr.ib(type=str, default=None)
    project = attr.ib(type=str, default=None)
    container = attr.ib(type=str, default=None)
    container_url = attr.ib(type=str, default=None)
    base_container = attr.ib(type=str, default=None)
    base_container_url = attr.ib(type=str, default=None)
    machine_type = attr.ib(type=str, default=None)
    cluster = attr.ib(type=str, default=None)
    cluster_id = attr.ib(type=str, default=None)
    usage_rate = attr.ib(type=str, default=None)
    started_by_user_id = attr.ib(type=str, default=None)
    parent_job_id = attr.ib(type=str, default=None)
    job_error = attr.ib(type=str, default=None)
    dt_created = attr.ib(type=str, default=None)
    dt_modified = attr.ib(type=str, default=None)
    dt_provisioning_started = attr.ib(type=str, default=None)
    dt_provisioning_finished = attr.ib(type=str, default=None)
    dt_started = attr.ib(type=str, default=None)
    dt_finished = attr.ib(type=str, default=None)
    dt_teardown_started = attr.ib(type=str, default=None)
    dt_teardown_finished = attr.ib(type=str, default=None)
    dt_deleted = attr.ib(type=str, default=None)
    exit_code = attr.ib(type=str, default=None)
    queue_position = attr.ib(type=str, default=None)
    seq_num = attr.ib(type=int, default=None)
    storage_region = attr.ib(type=str, default=None)
    cluster_machine = attr.ib(type=str, default=None)
    fqdn = attr.ib(type=str, default=None)
    ports = attr.ib(type=str, default=None)
    is_public = attr.ib(type=bool, default=None)
    container_user = attr.ib(type=str, default=None)
    has_code = attr.ib(type=bool, default=None)
    code_uploaded = attr.ib(type=bool, default=None)
    code_commit = attr.ib(type=str, default=None)
    run_till_cancelled = attr.ib(type=bool, default=None)
    push_on_completion = attr.ib(type=bool, default=None)
    new_image_name = attr.ib(type=str, default=None)
    cpu_hostname = attr.ib(type=str, default=None)
    cpu_count = attr.ib(type=int, default=None)
    cpu_model = attr.ib(type=str, default=None)
    cpu_flags = attr.ib(type=str, default=None)
    cpu_mem = attr.ib(type=str, default=None)
    gpu_name = attr.ib(type=str, default=None)
    gpu_serial = attr.ib(type=str, default=None)
    gpu_device = attr.ib(type=str, default=None)
    gpu_driver = attr.ib(type=str, default=None)
    gpu_count = attr.ib(type=int, default=None)
    gpu_mem = attr.ib(type=str, default=None)
    tpu_type = attr.ib(type=str, default=None)
    tpu_name = attr.ib(type=str, default=None)
    tpu_grpc_url = attr.ib(type=str, default=None)
    tpu_tf_version = attr.ib(type=str, default=None)
    tpu_dataset_dir = attr.ib(type=str, default=None)
    tpu_model_dir = attr.ib(type=str, default=None)
    target_node_attrs = attr.ib(type=dict, default=None)
    job_env = attr.ib(type=dict, default=None)
    env_vars = attr.ib(type=dict, default=None)
    shared_mem_mbytes = attr.ib(type=int, default=None)
    shutdown_timeout = attr.ib(type=int, default=None)
    is_preemptible = attr.ib(type=bool, default=None)
    metrics_url = attr.ib(type=str, default=None)
    custom_metrics = attr.ib(type=str, default=None)
    experiment_id = attr.ib(type=str, default=None)

    command = attr.ib(type=str, default=None)
    workspace_file_name = attr.ib(type=str, default=None)
    use_dockerfile = attr.ib(type=str, default=None)
    rel_dockerfile_path = attr.ib(type=str, default=None)
    registry_username = attr.ib(type=str, default=None)
    registry_password = attr.ib(type=str, default=None)
    build_only = attr.ib(type=bool, default=None)

    registry_target = attr.ib(type=str, default=None)
    registry_target_username = attr.ib(type=str, default=None)
    registry_target_password = attr.ib(type=str, default=None)
    tags = attr.ib(type=list, factory=list)
    datasets = attr.ib(type=list, factory=list)



import json

import click


class ChoiceType(click.Choice):
    """Takes a string-keyed map and converts cli-provided parameter to corresponding value"""

    def __init__(self, type_map, case_sensitive=True):
        super(ChoiceType, self).__init__(tuple(type_map.keys()), case_sensitive=case_sensitive)
        self.type_map = type_map

    def convert(self, value, param, ctx):
        value = super(ChoiceType, self).convert(value, param, ctx).upper()

        for key, val in self.type_map.items():
            if key.upper() == value:
                return val

        raise KeyError()


def json_string(val):
    """Wraps json.loads so the cli help shows proper option's type name instead of 'LOADS'"""
    if not isinstance(val, (dict, list)):
        val = json.loads(val)
    return val



import abc
import mimetypes
import multiprocessing
import os
import re
import threading
import uuid
import json
try:
    import queue
except ImportError:
    import Queue as queue
from xml.etree import ElementTree
from urllib.parse import urlparse

import halo
import requests
import six

from gradient import api_sdk
from gradient.api_sdk.sdk_exceptions import ResourceFetchingError
from gradient.api_sdk.utils import base64_encode
from gradient.cli_constants import CLI_PS_CLIENT_NAME
from gradient.cli.jobs import get_workspace_handler
from gradient.commands import jobs as jobs_commands
from gradient.commands.common import BaseCommand, DetailsCommandMixin, ListCommandPagerMixin
from gradient.commands.jobs import BaseCreateJobCommandMixin, BaseJobCommand, CreateJobCommand
from gradient.exceptions import ApplicationError

S3_XMLNS = 'http://s3.amazonaws.com/doc/2006-03-01/'
DATASET_IMPORTER_IMAGE = "paperspace/dataset-importer:latest"
PROJECT_NAME = "Job Builder"
SUPPORTED_URL = ['https', 'http']
IMPORTER_COMMAND = "go-getter"
HTTP_SECRET = "HTTP_AUTH"
S3_ACCESS_KEY = "AWS_ACCESS_KEY_ID"
S3_SECRET_KEY = "AWS_SECRET_ACCESS_KEY"
S3_REGION_KEY = "AWS_DEFAULT_REGION"


class WorkerPool(object):

    def __init__(self, count=None, min_count=4, max_count=16, cpu_multiplier=1):
        if count is None:
            count = min(max(round(multiprocessing.cpu_count() * cpu_multiplier), min_count), max_count)

        self._work = queue.Queue(maxsize=count)
        self._threads = [self._create_thread() for _ in range(count)]

        self._exception = None
        self._exception_lock = threading.Lock()

        self._completed_count = 0
        self._completed_lock = threading.Lock()

        self.worker_count = count

    def _create_thread(self):
        t = threading.Thread(target=self._worker)
        t.setDaemon(True)
        return t

    def __enter__(self):
        for thread in self._threads:
            thread.start()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._exception:
            raise self._exception

        for _ in range(self.worker_count):
            try:
                self._work.put(None, block=True, timeout=1)
            except queue.Full:
                break

        self._work.join()

        for thread in self._threads:
            thread.join()

    def _worker(self):
        while not self.has_exception():
            try:
                work = self._work.get(block=True, timeout=1)
            except queue.Empty:
                continue

            try:
                if work is None:
                    return

                (func, args, kwargs) = work
                func(*args, **kwargs)

                with self._completed_lock:
                    self._completed_count += 1
            except Exception as e:
                self.set_exception(e)
            finally:
                self._work.task_done()

    def put(self, func, *args, **kwargs):
        while not self.has_exception():
            try:
                return self._work.put((func, args, kwargs), block=True, timeout=1)
            except queue.Full:
                pass

    def set_exception(self, exception):
        with self._exception_lock:
            if not self._exception:
                self._exception = exception

    def has_exception(self):
        with self._exception_lock:
            return self._exception is not None

    def completed_count(self):
        with self._completed_lock:
            return self._completed_count


@six.add_metaclass(abc.ABCMeta)
class BaseDatasetsCommand(BaseCommand):
    def _get_client(self, api_key, logger):
        return api_sdk.clients.DatasetsClient(
            api_key=api_key,
            logger=logger,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )


class ListDatasetsCommand(ListCommandPagerMixin, BaseDatasetsCommand):
    def _get_table_data(self, objects):
        data = [('Name', 'ID', 'Storage Provider')]
        for dataset in objects:
            data.append((
                dataset.name,
                dataset.id,
                '{} ({})'.format(
                    dataset.storage_provider.name,
                    dataset.storage_provider.id
                ),
            ))
        return data


class ShowDatasetDetailsCommand(DetailsCommandMixin, BaseDatasetsCommand):
    def _get_table_data(self, instance):
        data = (
            ('Name', instance.name),
            ('ID', instance.id),
            ('Description', instance.description or ''),
            ('StorageProvider', '{} ({})'.format(
                instance.storage_provider.name,
                instance.storage_provider.id
            )),
        )
        return data


class CreateDatasetCommand(BaseDatasetsCommand):
    def execute(self, name, storage_provider_id, description=None):
        dataset_id = self.client.create(
            name=name,
            storage_provider_id=storage_provider_id,
            description=description,
        )
        self.logger.log('Created dataset: {}'.format(dataset_id))


class UpdateDatasetCommand(BaseDatasetsCommand):
    def execute(self, dataset_id, name=None, description=None):
        self.client.update(
            dataset_id=dataset_id,
            name=name,
            description=description,
        )
        self.logger.log('Updated dataset: {}'.format(dataset_id))


class DeleteDatasetCommand(BaseDatasetsCommand):
    def execute(self, dataset_id):
        self.client.delete(dataset_id)
        self.logger.log('Deleted dataset: {}'.format(dataset_id))


@six.add_metaclass(abc.ABCMeta)
class BaseDatasetVersionsCommand(BaseCommand):
    def _get_client(self, api_key, logger):
        return api_sdk.clients.DatasetVersionsClient(
            api_key=api_key,
            logger=logger,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )


class ListDatasetVersionsCommand(ListCommandPagerMixin, BaseDatasetVersionsCommand):
    def _get_table_data(self, objects):
        data = [('ID', 'Message', 'Tags')]
        for dataset_version in objects:
            data.append((
                '{}:{}'.format(self._id, dataset_version.version),
                dataset_version.message or '',
                ', '.join([t.name for t in dataset_version.tags])
            ))
        return data

    def execute(self, **kwargs):
        self._id = kwargs['dataset_id']
        return super(ListDatasetVersionsCommand, self).execute(**kwargs)


class ShowDatasetVersionDetailsCommand(DetailsCommandMixin, BaseDatasetVersionsCommand):
    def _get_table_data(self, instance):
        data = (
            ('ID', self._id),
            ('Message', instance.message or ''),
            ('Committed', str(instance.is_committed).lower()),
            ('Tags', ', '.join([t.name for t in instance.tags])),
        )
        return data

    def execute(self, dataset_version_id, **kwargs):
        self._id = dataset_version_id
        return super(ShowDatasetVersionDetailsCommand, self).execute(dataset_version_id, **kwargs)


class CreateDatasetVersionCommand(BaseDatasetVersionsCommand):
    def execute(self, dataset_id, message=None, source_paths=None):
        if source_paths:
            for source_path in source_paths:
                if not os.path.exists(source_path):
                    raise ApplicationError('Source path not found: {}'.format(source_path))

        version = self.client.create(dataset_id=dataset_id, message=message)
        dataset_version_id = '{}:{}'.format(dataset_id, version)
        self.logger.log('Created dataset version: {}'.format(dataset_version_id))

        if source_paths:
            create = PutDatasetFilesCommand(api_key=self.api_key, logger=self.logger)
            create.execute(dataset_version_id, source_paths=source_paths, target_path='/')

            commit = CommitDatasetVersionCommand(api_key=self.api_key, logger=self.logger)
            commit.execute(dataset_version_id)


class UpdateDatasetVersionCommand(BaseDatasetVersionsCommand):
    def execute(self, dataset_version_id, message=None):
        self.client.update(dataset_version_id, message=message)
        self.logger.log('Updated dataset version: {}'.format(dataset_version_id))


class CommitDatasetVersionCommand(BaseDatasetVersionsCommand):
    def execute(self, dataset_version_id, message=None):
        dataset_version = self.client.get(dataset_version_id)
        if dataset_version.is_committed:
            self.logger.log('Dataset version already committed')
            return

        self.client.update(dataset_version_id, is_committed=True)
        self.logger.log('Committed dataset version: {}'.format(dataset_version_id))


class DeleteDatasetVersionCommand(BaseDatasetVersionsCommand):
    def execute(self, dataset_version_id):
        self.client.delete(dataset_version_id)
        self.logger.log('Deleted dataset version: {}'.format(dataset_version_id))


@six.add_metaclass(abc.ABCMeta)
class BaseDatasetTagsCommand(BaseCommand):
    def _get_client(self, api_key, logger):
        return api_sdk.clients.DatasetTagsClient(
            api_key=api_key,
            logger=logger,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )


class ListDatasetTagsCommand(ListCommandPagerMixin, BaseDatasetTagsCommand):
    def _get_table_data(self, objects):
        data = [('ID', 'Name')]
        for dataset_tag in objects:
            data.append((
                '{}:{}'.format(self._id, dataset_tag.version.version),
                dataset_tag.name,
            ))
        return data

    def execute(self, **kwargs):
        self._id, _, _ = kwargs['dataset_id'].partition(':')
        return super(ListDatasetTagsCommand, self).execute(**kwargs)


class SetDatasetTagCommand(BaseDatasetTagsCommand):
    def execute(self, dataset_version_id, name):
        self.client.set(dataset_version_id, name=name)
        dataset_id, _, _ = dataset_version_id.partition(':')
        self.logger.log('Set dataset tag: {}:{}'.format(dataset_id, name))


class DeleteDatasetTagCommand(BaseDatasetTagsCommand):
    def execute(self, dataset_tag_id):
        self.client.delete(dataset_tag_id)
        self.logger.log('Deleted dataset tag: {}'.format(dataset_tag_id))


@six.add_metaclass(abc.ABCMeta)
class BaseDatasetFilesCommand(BaseDatasetVersionsCommand):
    def __init__(self, *args, **kwargs):
        super(BaseDatasetFilesCommand, self).__init__(*args, **kwargs)
        self.dataset_client = api_sdk.clients.DatasetsClient(
            api_key=self.api_key,
            logger=self.logger,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )

    def assert_supported(self, dataset_id):
        dataset_id, _, _ = dataset_id.partition(':')

        dataset = self.dataset_client.get(dataset_id)
        if dataset.storage_provider.type != 's3':
            raise ApplicationError('%s storage type not supported' % dataset.storage_provider.type)

    def resolve_dataset_version_id(self, dataset_ref):
        try:
            dataset_ref = self.dataset_client.get_ref(dataset_ref)
            if dataset_ref and dataset_ref.version:
                dataset_ref = '{}:{}'.format(dataset_ref.id, dataset_ref.version.version)
        except ResourceFetchingError:
            pass
        return dataset_ref

    @staticmethod
    def validate_s3_response(response):
        if not response.ok:
            raise ApplicationError('Failed to execute request against storage provider: %s\n\n%s' %
                                   (response.status_code, response.text))

    @staticmethod
    def report_connection_error(exception):
        raise ApplicationError('Failed to execute request against storage provider: %s' %
                               exception)

    @staticmethod
    def normalize_path(path):
        if not path:
            return '/'
        path = re.sub(r'/+', '/', path)
        if not path.startswith('/'):
            path = '/' + path
        return path

    def get_object(self, dataset_version_id, path):
        path = path.lstrip('/')

        if not path:
            return

        pre_signed = self.client.generate_pre_signed_s3_url(
            dataset_version_id,
            method='headObject',
            params={'Key': path},
        )

        try:
            response = requests.head(pre_signed.url)
            if response.status_code == 404:
                return
            self.validate_s3_response(response)

            size = response.headers.get('Content-Length', 0)
            return {'key': path, 'size': size}
        except requests.exceptions.ConnectionError as e:
            return self.report_connection_error(e)

    def list_objects(self, dataset_version_id, recursive=False, path='/', absolute=False, max_keys=20):
        path = self.normalize_path(path)

        if not path.endswith('/'):
            path += '/'

        next_continuation_token = None

        while True:
            params = {'Prefix': path, 'MaxKeys': max_keys}
            if next_continuation_token:
                params['ContinuationToken'] = next_continuation_token
            if recursive:
                params['Delimiter'] = ''

            pre_signed = self.client.generate_pre_signed_s3_url(
                dataset_version_id,
                method='listObjectsV2',
                params=params,
            )

            try:
                response = requests.get(pre_signed.url)
                self.validate_s3_response(response)
            except requests.exceptions.ConnectionError as e:
                self.report_connection_error(e)
                return

            tree = ElementTree.fromstring(response.text)

            prefix = tree.find('{' + S3_XMLNS + '}Prefix').text
            results = []
            next_continuation_token = None

            key_prefix = path[1:] if absolute else ''

            for item in tree:
                name = item.tag.rpartition('}')[2]
                if name == 'Contents':
                    key = item.find('{' + S3_XMLNS + '}Key').text[len(prefix):]
                    is_dir = key.endswith('/')

                    if not key or (recursive and is_dir):
                        continue

                    result = {'key': key_prefix + key}
                    if not is_dir:
                        result['size'] = item.find('{' + S3_XMLNS + '}Size').text

                    results.append(result)
                elif name == 'NextContinuationToken':
                    next_continuation_token = item.text
                elif name == 'CommonPrefixes':
                    if recursive:
                        continue
                    key = item.find('{' + S3_XMLNS + '}Prefix').text[len(prefix):]
                    results.append({'key': key_prefix + key})

            yield results, bool(next_continuation_token)

            if not next_continuation_token:
                break


class ListDatasetFilesCommand(ListCommandPagerMixin, BaseDatasetFilesCommand):
    def _get_table_data(self, objects):
        data = [('Name', 'Size')]
        for obj in objects:
            data.append((
                obj['key'],
                obj.get('size', ''),
            ))
        return data

    def _get_instances(self, kwargs):
        self.assert_supported(kwargs['dataset_version_id'])

        kwargs['dataset_version_id'] = self.resolve_dataset_version_id(kwargs['dataset_version_id'])

        return self.list_objects(**kwargs)


class GetDatasetFilesCommand(BaseDatasetFilesCommand):

    @classmethod
    def _get(cls, url, path):
        dir_path = os.path.dirname(path)
        tmp_path = path + '.tmp-%s' % uuid.uuid4()

        if os.path.exists(path) and not os.path.isfile(path):
            raise ApplicationError('%s already exists' % path)

        os.makedirs(dir_path, exist_ok=True)

        try:
            with requests.Session() as session:
                try:
                    with session.get(url, stream=True) as r:
                        cls.validate_s3_response(r)
                        with open(tmp_path, 'wb') as f:
                            for chunk in r.iter_content(chunk_size=8192):
                                f.write(chunk)
                except requests.exceptions.ConnectionError as e:
                    return cls.report_connection_error(e)

            os.rename(tmp_path, path)
        finally:
            if os.path.isfile(tmp_path):
                os.remove(tmp_path)

    def execute(self, dataset_version_id, source_paths, target_path):
        self.assert_supported(dataset_version_id)

        dataset_version_id = self.resolve_dataset_version_id(dataset_version_id)

        target_path = os.path.abspath(target_path)

        if not source_paths:
            source_paths = ['/']

        status_text = 'Downloading files'

        with halo.Halo(text=status_text, spinner='dots') as status:
            with WorkerPool() as pool:
                for source_path in source_paths:
                    source_path = self.normalize_path(source_path)

                    list_objects = None
                    is_file = False
                    has_trailing_slash = source_path.endswith('/')

                    if not has_trailing_slash:
                        result = self.get_object(dataset_version_id, source_path)
                        if result is not None:
                            list_objects = [([result], False)]
                            is_file = True

                    if not list_objects:
                        list_objects = self.list_objects(
                            dataset_version_id=dataset_version_id,
                            path=source_path,
                            recursive=True,
                            absolute=True,
                            max_keys=max(pool.worker_count * 2, 64)
                        )

                    def update_status():
                        status.text = '{}: {} ({})  '.format(status_text, source_path, pool.completed_count())

                    for results, _ in list_objects:
                        if not results:
                            break

                        pre_signeds = self.client.generate_pre_signed_s3_urls(
                            dataset_version_id,
                            calls=[dict(method='getObject', params=dict(Key=r['key'])) for r in results],
                        )

                        for result, pre_signed in zip(results, pre_signeds):
                            if is_file:
                                path = target_path
                            elif has_trailing_slash:
                                path = os.path.join(target_path, result['key'][len(source_path)-1:])
                            else:
                                path = os.path.join(target_path, result['key'])

                            update_status()
                            pool.put(self._get, url=pre_signed.url, path=path)


class PutDatasetFilesCommand(BaseDatasetFilesCommand):

    @classmethod
    def _put(cls, path, url, content_type):
        size = os.path.getsize(path)

        with requests.Session() as session:
            headers = {'Content-Type': content_type}

            try:
                if size > 0:
                    with open(path, 'rb') as f:
                        r = session.put(url, data=f, headers=headers, timeout=5)
                else:
                    headers.update({'Content-Size': '0'})
                    r = session.put(url, data='', headers=headers, timeout=5)

                cls.validate_s3_response(r)
            except requests.exceptions.ConnectionError as e:
                return cls.report_connection_error(e)

    @staticmethod
    def _list_files(source_path):
        if os.path.isfile(source_path):
            yield True, source_path
            return

        if os.path.isdir(source_path):
            for dir_path, _, names in os.walk(source_path):
                for name in names:
                    yield False, os.path.join(dir_path, name)
            return

        raise ApplicationError('Invalid source path: ' + source_path)

    def _sign_and_put(self, dataset_version_id, pool, results, update_status):
        pre_signeds = self.client.generate_pre_signed_s3_urls(
            dataset_version_id,
            calls=[dict(method='putObject', params=dict(Key=r['key'], ContentType=r['mimetype'])) for r in results],
        )

        for pre_signed, result in zip(pre_signeds, results):
            update_status()
            pool.put(self._put, url=pre_signed.url, path=result['path'], content_type=result['mimetype'])

    def execute(self, dataset_version_id, source_paths, target_path):
        self.assert_supported(dataset_version_id)

        if not target_path:
            target_path = '/'
        else:
            target_path = self.normalize_path(target_path)
            if not target_path.endswith('/'):
                target_path += '/'

        status_text = 'Uploading files'

        with halo.Halo(text=status_text, spinner='dots') as status:
            with WorkerPool() as pool:
                for source_path in source_paths:
                    has_trailing_slash = source_path.endswith(os.path.sep)
                    source_path = os.path.abspath(source_path)
                    source_name = os.path.basename(source_path)

                    def update_status():
                        status.text = '{}: {} ({})'.format(status_text, source_path, pool.completed_count())

                    results = []

                    for source_path_is_file, path in self._list_files(source_path):
                        path = path.replace(os.path.sep, '/')

                        key = target_path
                        if not source_path_is_file:
                            if not has_trailing_slash:
                                key += source_name + '/'
                            key += path[len(source_path)+1:]

                        mimetype = mimetypes.guess_type(key)[0] or 'application/octet-stream'

                        results.append(dict(key=key, path=path, mimetype=mimetype))

                        if len(results) == pool.worker_count:
                            self._sign_and_put(dataset_version_id, pool, results, update_status)
                            results = []

                    if results:
                        self._sign_and_put(dataset_version_id, pool, results, update_status)


class DeleteDatasetFilesCommand(BaseDatasetFilesCommand):

    @classmethod
    def _delete(cls, url):
        with requests.Session() as session:
            try:
                r = session.delete(url)
                cls.validate_s3_response(r)
            except requests.exceptions.ConnectionError as e:
                return cls.report_connection_error(e)

    def execute(self, dataset_version_id, paths):
        self.assert_supported(dataset_version_id)

        status_text = 'Deleting files'

        with halo.Halo(text=status_text, spinner='dots') as status:
            with WorkerPool() as pool:
                for path in paths:
                    path = self.normalize_path(path)

                    list_objects = None
                    has_trailing_slash = path.endswith('/')

                    def update_status():
                        status.text = '{}: {} ({})'.format(status_text, path, pool.completed_count())

                    if not has_trailing_slash:
                        result = self.get_object(dataset_version_id, path)
                        if result is not None:
                            list_objects = [([result], False)]

                    if not list_objects:
                        list_objects = self.list_objects(
                            dataset_version_id=dataset_version_id,
                            path=path,
                            recursive=True,
                            absolute=True,
                        )

                    for results, _ in list_objects:
                        if not results:
                            break

                        pre_signeds = self.client.generate_pre_signed_s3_urls(
                            dataset_version_id,
                            calls=[dict(method='deleteObject', params=dict(Key=r['key'])) for r in results],
                        )

                        for pre_signed in pre_signeds:
                            update_status()
                            pool.put(self._delete, url=pre_signed.url)


class ImportDatasetCommand(BaseCreateJobCommandMixin):
    def create_secret(self, key, value, expires_in=86400):
        client = api_sdk.clients.SecretsClient(
            api_key=self.api_key,
            logger=self.logger,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )

        response = client.ephemeral(key, value, expires_in)
        return response

    def get_command(self, s3_url, http_url, http_auth):
        command = "%s %s /data/output" % (IMPORTER_COMMAND, (s3_url or http_url))
        if s3_url:
            command = "%s s3::%s /data/output" % (IMPORTER_COMMAND, s3_url)

        if http_url and http_auth is not None:
            url = urlparse(http_url)
            command_string = "%s https://${{HTTP_AUTH}}@%s /data/output" % (IMPORTER_COMMAND, url.path)
            command = base64_encode(command_string)

        return command
    
    def get_env_vars(self, s3_url, http_url, secrets):
        if s3_url is not None:
            if secrets[S3_ACCESS_KEY] is None or secrets[S3_SECRET_KEY] is None:
                self.logger.log('s3AccessKey and s3SecretKey required')
                return 

            access_key_secret = self.create_secret(S3_ACCESS_KEY, secrets[S3_ACCESS_KEY])
            secret_key_secret = self.create_secret(S3_SECRET_KEY, secrets[S3_SECRET_KEY])

            access_key_value = "secret:ephemeral:%s" % access_key_secret[S3_ACCESS_KEY]
            secret_key_value = "secret:ephemeral:%s" % secret_key_secret[S3_SECRET_KEY]

            return {
                S3_ACCESS_KEY: access_key_value,
                S3_SECRET_KEY: secret_key_value,
            }

        if http_url and secrets[S3_ACCESS_KEY] is not None:
            http_auth_secret = self.create_secret(HTTP_SECRET, secrets[HTTP_SECRET])
            return {
                HTTP_SECRET: http_auth_secret
            }
        
        return ""

    def _create(self, workflow):
        client = api_sdk.clients.JobsClient(
            api_key=self.api_key,
            ps_client_name=CLI_PS_CLIENT_NAME,
        )
        return self.client.create(**workflow)


    def execute(self, cluster_id, machine_type, dataset_id, s3_url, http_url, http_auth, access_key, secret_key):
        if s3_url is None and http_url is None:
            self.logger.log('Error: --s3Url or --httpUrl required')
            return

        workflow = {
            "cluster_id": cluster_id,
            "container": DATASET_IMPORTER_IMAGE,
            "machine_type": machine_type,
            "project": PROJECT_NAME,
            "datasets": [{ "id": dataset_id, "name": "output", "output": True }],
            "project_id": None
        }

        dataset_url = s3_url or http_url

        url = urlparse(dataset_url)
        if url.scheme not in SUPPORTED_URL:
            self.logger.log('Invalid URL format supported [{}] Format:{} URL:{}'.format(','.join(SUPPORTED_URL), url.scheme, dataset_url))
            return

        command = self.get_command(s3_url, http_url, http_auth)
        if command:
            workflow["command"] = command

        env_vars = self.get_env_vars(s3_url, http_url, { HTTP_SECRET: http_auth, S3_ACCESS_KEY: access_key, S3_SECRET_KEY: secret_key })
        if env_vars:
            workflow["env_vars"] = env_vars

        command = CreateJobCommand(api_key=self.api_key, workspace_handler=get_workspace_handler())
        command.execute(workflow)



import json

from .base_client import BaseClient, TagsSupportMixin
from .. import repositories, models


class ModelsClient(TagsSupportMixin, BaseClient):
    entity = "mlModel"

    def list(self, experiment_id=None, project_id=None, tags=None):
        """Get list of models

        :param str experiment_id: Experiment ID to filter models
        :param str project_id: Project ID to filter models
        :param list[str]|tuple[str] tags: tags to filter models

        :returns: List of Model instances
        :rtype: list[models.Model]
        """
        repository = self.build_repository(repositories.ListModels)
        models_list = repository.list(experiment_id=experiment_id, project_id=project_id, tags=tags)
        return models_list

    def delete(self, model_id):
        """Delete a model

        :param str model_id: Model ID
        """
        repository = self.build_repository(repositories.DeleteModel)
        repository.delete(model_id)

    def upload(self, path, name, model_type, model_summary=None, notes=None, tags=None, project_id=None, cluster_id=None):
        """Upload model

        :param file path: path to Model
        :param str name: Model name
        :param str model_type: Model Type
        :param dict|None model_summary: Dictionary describing model parameters like loss, accuracy, etc.
        :param str|None notes: Optional model description
        :param list[str] tags: List of tags
        :param str|None project_id: ID of a project
        :param str|None cluster_id: ID of a cluster

        :return: ID of new model
        :rtype: str
        """

        model = models.Model(
            name=name,
            model_type=model_type,
            summary=json.dumps(model_summary) if model_summary else None,
            notes=notes,
            project_id=project_id,
        )

        repository = self.build_repository(repositories.UploadModel)
        model_id = repository.create(model, path=path, cluster_id=cluster_id)

        if tags:
            self.add_tags(entity_id=model_id, tags=tags)

        return model_id

    def create(self, name, model_type, dataset_ref, model_summary=None, notes=None, tags=None, project_id=None):
        """Create model

        :param str name: Model name
        :param str model_type: Model Type
        :param str dataset_ref: Dataset ref to associate a model with
        :param dict|None model_summary: Dictionary describing model parameters like loss, accuracy, etc.
        :param str|None notes: Optional model description
        :param list[str] tags: List of tags
        :param str|None project_id: ID of a project

        :return: ID of new model
        :rtype: str
        """

        model = models.Model(
            name=name,
            model_type=model_type,
            dataset_ref=dataset_ref,
            summary=json.dumps(model_summary) if model_summary else None,
            notes=notes,
            project_id=project_id,
        )

        repository = self.build_repository(repositories.CreateModel)
        model_id = repository.create(model)

        if tags:
            self.add_tags(entity_id=model_id, tags=tags)

        return model_id

    def get(self, model_id):
        """Get model instance

        :param str model_id:
        :return: Model instance
        :rtype: models.Model
        """
        repository = self.build_repository(repositories.GetModel)
        model = repository.get(model_id=model_id)
        return model

    def get_model_files(self, model_id, links=False, size=False):
        """Get list of models

        :param str model_id: Model ID
        :param bool links: Get links to model files
        :param bool size: Get sizes of each file in bytes

        :rtype: list[models.ModelFile]
        """
        repository = self.build_repository(repositories.ListModelFiles)
        models_list = repository.list(model_id=model_id, links=links, size=size)
        return models_list



import time

from .common import BaseRepository, CreateResource, DeleteResource, ListResources, StartResource, StopResource, \
    GetResource, AlterResource
from .. import serializers, models
from ..config import config
from ..serializers import MachineSchema


class MachinesApiUrlMixin(object):
    def _get_api_url(self, **kwargs):
        return config.CONFIG_HOST


class CheckMachineAvailability(MachinesApiUrlMixin, BaseRepository):
    def get_request_url(self, **kwargs):
        return "/machines/getAvailability/"

    def get(self, machine_type, region):
        kwargs = {"machineType": machine_type,
                  "region": region}
        response = self._get(**kwargs)
        self._validate_response(response)
        is_available = response.data["available"]
        return is_available

    def _get_request_params(self, kwargs):
        return kwargs


class CreateMachine(MachinesApiUrlMixin, CreateResource):
    SERIALIZER_CLS = serializers.MachineSchema
    HANDLE_FIELD = "id"

    def get_request_url(self, **kwargs):
        return "/machines/createSingleMachinePublic/"


class DeleteMachine(MachinesApiUrlMixin, DeleteResource):
    def get_request_url(self, **kwargs):
        machine_id = kwargs["id"]
        url = "/machines/{}/destroyMachine/".format(machine_id)
        return url

    def _send_request(self, client, url, json_data=None):
        response = client.post(url, json=json_data)
        return response

    def _get_request_json(self, kwargs):
        release_public_ip = kwargs["release_public_ip"]
        json_ = {"releasePublicIp": release_public_ip} if release_public_ip else None
        return json_


class ListMachines(MachinesApiUrlMixin, ListResources):
    def _parse_objects(self, data, **kwargs):
        instances = []

        for machine_dict in data:
            machine = serializers.MachineSchema().get_instance(machine_dict)
            instances.append(machine)

        return instances

    def get_request_url(self, **kwargs):
        return "/machines/getMachines/"

    def _get_request_json(self, kwargs):
        instance = models.Machine(**kwargs)
        serializer = serializers.MachineSchemaForListing()
        marshaled = serializer.dump(instance)
        instance_dict = marshaled.data
        instance_dict_without_nulls = {key: val for key, val in instance_dict.items() if val is not None}
        json_ = {"params": instance_dict_without_nulls} if instance_dict_without_nulls else None
        return json_


class RestartMachine(MachinesApiUrlMixin, StartResource):
    VALIDATION_ERROR_MESSAGE = "Unable to restart instance"

    def restart(self, id_, **kwargs):
        self._run(id=id_, **kwargs)

    def get_request_url(self, **kwargs):
        machine_id = kwargs["id"]
        url = "/machines/{}/restart/".format(machine_id)
        return url

    def _send_request(self, client, url, json_data=None):
        response = client.post(url, json=json_data)
        return response


class StartMachine(MachinesApiUrlMixin, StartResource):
    def restart(self, id_, **kwargs):
        self._run(id=id_, **kwargs)

    def get_request_url(self, **kwargs):
        machine_id = kwargs["id"]
        url = "/machines/{}/start/".format(machine_id)
        return url

    def _send_request(self, client, url, json_data=None):
        response = client.post(url, json=json_data)
        return response


class StopMachine(MachinesApiUrlMixin, StopResource):
    def restart(self, id_):
        self._run(id=id_)

    def get_request_url(self, **kwargs):
        machine_id = kwargs["id"]
        url = "/machines/{}/stop/".format(machine_id)
        return url

    def _send_request(self, client, url, json_data=None):
        response = client.post(url, json=json_data)
        return response


class GetMachine(MachinesApiUrlMixin, GetResource):
    SERIALIZER_CLS = MachineSchema

    def get_request_url(self, **kwargs):
        return "/machines/getMachinePublic/"

    def _get_request_params(self, kwargs):
        machine_id = kwargs["id"]
        params = {"machineId": machine_id}
        return params


class UpdateMachine(MachinesApiUrlMixin, AlterResource):
    SERIALIZER_CLS = serializers.MachineSchema
    VALIDATION_ERROR_MESSAGE = "Failed to update resource"

    def get_request_url(self, **kwargs):
        machine_id = kwargs["id"]
        url = "/machines/{}/updateMachinePublic/".format(machine_id)
        return url

    def _get_request_json(self, kwargs):
        kwargs.pop("id", None)
        return kwargs


class GetMachineUtilization(MachinesApiUrlMixin, GetResource):
    def _parse_object(self, data, **kwargs):
        instance = models.MachineUtilization(
            machine_id=data["machineId"],
            machine_seconds_used=data["utilization"].get("secondsUsed") if "utilization" in data else None,
            machine_billing_month=data["utilization"].get("billingMonth") if "utilization" in data else None,
            machine_hourly_rate=data["utilization"].get("hourlyRate") if "utilization" in data else None,
            storage_seconds_used=data["storageUtilization"].get("secondsUsed") if "utilization" in data else None,
            storage_billing_month=data["storageUtilization"].get("billingMonth") if "utilization" in data else None,
            storage_monthly_rate=data["storageUtilization"].get("monthlyRate") if "utilization" in data else None,
        )
        return instance

    def get_request_url(self, **kwargs):
        return "machines/getUtilization/"

    def _get_request_params(self, kwargs):
        d = {
            "machineId": kwargs["id"],
            "billingMonth": kwargs["billing_month"]
        }
        return d


class WaitForState(object):
    def __init__(self, api_key, logger, ps_client_name=None):
        self.api_key = api_key
        self.logger = logger
        self.get_machine_repository = GetMachine(api_key=api_key, logger=logger, ps_client_name=ps_client_name)

    def wait_for_state(self, machine_id, state, interval=5):

        while True:
            current_state = self._get_machine_state(machine_id)
            if current_state == state:
                return

            time.sleep(interval)

    def _get_machine_state(self, machine_id):
        machine = self.get_machine_repository.get(id=machine_id)
        return machine.state
